import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.karma.views import KarmaListView
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting
from pyplanet.apps.core.statistics.models import Score
from pyplanet.apps.core.maniaplanet.models import Player

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.apps.contrib.karma.views import KarmaWidget
from pyplanet.apps.contrib.karma.mxkarma import MXKarma

from .models import Karma as KarmaModel


class Karma(AppConfig):
	name = 'pyplanet.apps.contrib.karma'
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.current_votes = []
		self.current_karma = 0.0
		self.current_karma_percentage = 0.0
		self.current_karma_positive = 0.0
		self.current_karma_negative = 0.0
		self.widget = None

		self.setting_expanded_voting = Setting(
			'expanded_voting', 'Expanded voting', Setting.CAT_BEHAVIOUR, type=bool,
			description='Expands the karma voting to also use ---, -, +-, + and +++.',
			default=False
		)

		self.setting_finishes_before_voting = Setting(
			'finishes_before_voting', 'Finishes before voting', Setting.CAT_BEHAVIOUR, type=int,
			description='Amount of times a player should finish before being allowed to vote.',
			default=0
		)

		self.mx_karma = MXKarma(self)

	async def on_start(self):
		# Register commands.
		await self.instance.command_manager.register(Command(command='whokarma', target=self.show_map_list,
															 description='Displays who voted what on the current map.'))

		# Register signals.
		self.context.signals.listen(mp_signals.map.map_begin, self.map_begin)
		self.context.signals.listen(mp_signals.map.map_end, self.mx_karma.map_end)
		self.context.signals.listen(mp_signals.player.player_chat, self.player_chat)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		await self.context.setting.register(self.setting_finishes_before_voting, self.setting_expanded_voting)

		# Load initial data.
		await self.get_votes_list(self.instance.map_manager.current_map)
		await self.calculate_karma()
		await self.mx_karma.on_start()

		await self.chat_current_karma()

		self.widget = KarmaWidget(self)
		await self.widget.display()

		await self.load_map_votes()

	async def on_stop(self):
		await self.mx_karma.on_stop()

	async def load_map_votes(self, map=None):
		if map:
			map.karma = await self.get_map_karma(map)
		else:
			maps = {m.id: m for m in self.instance.map_manager.maps}
			map_karmas = dict()

			# Fetch all.
			rows = await KarmaModel.execute(
				KarmaModel.select().where(
					KarmaModel.map_id << list(maps.keys())
				)
			)

			# Group by map.
			for row in rows:
				if row.map_id not in map_karmas:
					map_karmas[row.map_id] = list()
				map_karmas[row.map_id].append(row)

			# Map karma stats.
			for map_id, karma_list in map_karmas.items():
				total_score = 0.0
				for vote in karma_list:
					if vote.expanded_score is not None:
						total_score += vote.expanded_score
					else:
						total_score += vote.score

				maps[map_id].karma = dict(
					vote_count=len(karma_list),
					map_karma=total_score
				)

			del map_karmas
			del maps

	async def show_map_list(self, player, map=None, **kwargs):
		"""
		Show map list to player for current map or map provided.. Provide player instance.

		:param player: Player instance.
		:param map: Map instance or current map.
		:param kwargs: ...
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: View instance.
		"""
		view = KarmaListView(self, map or self.instance.map_manager.current_map)
		await view.display(player=player.login)
		return view

	async def map_begin(self, map):
		await self.get_votes_list(map)
		await self.calculate_karma()
		await self.mx_karma.map_begin(map)
		await self.chat_current_karma()

		await self.widget.display()

	async def player_connect(self, player, is_spectator, source, signal):
		await self.mx_karma.player_connect(player=player)
		await self.widget.display(player=player)

	async def player_chat(self, player, text, cmd):
		if not cmd:
			if text == '+++' or text == '++' or text == '+' or text == '+-' or text == '-+' or text == '-' or text == '--' or text == '---':
				expanded_voting = await self.setting_expanded_voting.get_value()
				if expanded_voting is False:
					if text == '+++' or text == '+' or text == '+-' or text == '-+' or text == '-' or text == '---':
						return

				if self.instance.game.game == 'tm' or self.instance.game.game == 'tmnext':
					finishes_required = await self.setting_finishes_before_voting.get_value()
					player_finishes = await Score.objects.count(Score.select().where(Score.map_id == self.instance.map_manager.current_map.get_id()).where(Score.player_id == player.get_id()))
					if player_finishes < finishes_required:
						message = '$i$f00You have to finish this map at least $fff{}$f00 times before voting!'.format(finishes_required)
						await self.instance.chat(message, player)
						return

				normal_score = -1
				score = -1
				if text == '++' or text == '+++':
					normal_score = 1
					score = 1
				elif text == '+':
					normal_score = 1
					score = 0.5
				elif text == '+-' or text == '-+':
					score = 0
				elif text == '-':
					score = -0.5

				player_votes = [x for x in self.current_votes if x.player_id == player.get_id()]
				if len(player_votes) > 0:
					player_vote = player_votes[0]
					if (player_vote.expanded_score is not None and player_vote.expanded_score != score) or \
						(player_vote.expanded_score is None and player_vote.score != score):
						player_vote.score = normal_score
						player_vote.expanded_score = score
						await player_vote.save()

						map = next((m for m in self.instance.map_manager.maps if m.uid == self.instance.map_manager.current_map.uid), None)
						if map is not None:
							map.karma = await self.get_map_karma(self.instance.map_manager.current_map)

						message = '$ff0Successfully changed your karma vote to $fff{}$ff0{}!'.format(text,
							(' (same as $fff{}$ff0)'.format(text[:2]) if text == '+++' or text == '---' else '')
						)
						await self.calculate_karma()
						await asyncio.gather(
							self.instance.chat(message, player),
							self.widget.display()
						)
					else:
						message = '$ff0You have already voted $fff{}$ff0 on this map!'.format(text)
						await self.instance.chat(message, player)
				else:
					new_vote = KarmaModel(map=self.instance.map_manager.current_map, player=player, score=normal_score, expanded_score=score)
					await new_vote.save()

					self.current_votes.append(new_vote)
					await self.calculate_karma()

					map = next((m for m in self.instance.map_manager.maps if m.uid == self.instance.map_manager.current_map.uid), None)
					if map is not None:
						map.karma = await self.get_map_karma(self.instance.map_manager.current_map)

					message = '$ff0Successfully voted $fff{}$ff0{}!'.format(text,
						(' (same as $fff{}$ff0)'.format(text[:2]) if text == '+++' or text == '---' else '')
					)
					await asyncio.gather(
						self.instance.chat(message, player),
						self.widget.display()
					)

				# Reload map referenced information
				asyncio.ensure_future(self.load_map_votes(map=self.instance.map_manager.current_map))

	async def get_map_karma(self, map):
		vote_list = await KarmaModel.objects.execute(KarmaModel.select().where(KarmaModel.map_id == map.get_id()))

		total_score = 0.0
		for vote in vote_list:
			if vote.expanded_score is not None:
				total_score += vote.expanded_score
			else:
				total_score += vote.score

		return dict(
			vote_count=len(vote_list),
			map_karma=total_score
		)

	async def get_votes_list(self, map):
		vote_list = await KarmaModel.objects.execute(KarmaModel.select(KarmaModel, Player).join(Player).where(KarmaModel.map_id == map.get_id()))
		self.current_votes = list(vote_list)

	async def calculate_karma(self):
		total_score = 0.0
		total_abs = 0.0
		self.current_karma_positive = 0.0
		self.current_karma_negative = 0.0

		for vote in self.current_votes:
			score = vote.score
			if vote.expanded_score is not None:
				score = vote.expanded_score

			total_score += score
			total_abs += abs(score)
			if score > 0:
				self.current_karma_positive += score

		self.current_karma_negative = (total_abs - self.current_karma_positive)
		self.current_karma = total_score
		self.current_karma_percentage = 0
		if self.current_karma_positive > 0:
			self.current_karma_percentage = (self.current_karma_positive / total_abs)

	async def chat_current_karma(self):
		mx_karma = ''
		if self.mx_karma.api.activated:
			mx_karma = ', MX: $fff{}%$ff0 [$fff{}$ff0 votes]'.format(round(self.mx_karma.current_average, 1), self.mx_karma.current_count)

		num_current_votes = len(self.current_votes)
		message = '$ff0Current map karma: $fff{}$ff0 ($fff{}%$ff0) [$fff{}$ff0 votes]{}'.format(
			self.current_karma, round(self.current_karma_percentage * 100, 2), num_current_votes, mx_karma
		)
		await self.instance.chat(message)



from pyplanet.utils.times import format_time
from pyplanet.views.generics import ManualListView


class TopActivePlayersView(ManualListView):
	title = 'Top active players on the server'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, player, active):
		"""
		Init top active players list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.provide_search = False
		self.active = active

	async def get_data(self):
		data = list()
		for idx, player in enumerate(self.active):
			data.append(dict(
				place=idx+1,
				player_nickname=player.nickname,
				total_time=format_time(
					player.total_playtime * 1000, hide_hours_when_zero=False, hide_milliseconds=True
				),
			))

		return data

	async def destroy(self):
		self.active = None
		await super().destroy()

	def destroy_sync(self):
		self.active = None
		super().destroy_sync()

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'place',
				'sorting': False,
				'searching': False,
				'width': 6,
				'type': 'label'
			},
			{
				'name': 'Player',
				'index': 'player_nickname',
				'sorting': False,
				'searching': False,
				'width': 80,
				'type': 'label'
			},
			{
				'name': 'Total time on server',
				'index': 'total_time',
				'sorting': True,
				'searching': False,
				'width': 60,
				'type': 'label'
			},
		]


class TopDonatingPlayersView(ManualListView):
	title = 'Top donating players on the server'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Statistics'

	def __init__(self, app, player, donators):
		"""
		Init top donating players list view.

		:param player: Player instance.
		:param app: App instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		super().__init__(self)

		self.app = app
		self.player = player
		self.manager = app.context.ui
		self.provide_search = False
		self.donators = donators

	async def get_data(self):
		return [dict(
			place=idx+1,
			player_nickname=player.nickname,
			total_donations=player.total_donations
		) for idx, player in enumerate(self.donators)]

	async def destroy(self):
		self.donators = None
		await super().destroy()

	def destroy_sync(self):
		self.donators = None
		super().destroy_sync()

	async def get_fields(self):
		return [
			{
				'name': '#',
				'index': 'place',
				'sorting': False,
				'searching': False,
				'width': 6,
				'type': 'label'
			},
			{
				'name': 'Player',
				'index': 'player_nickname',
				'sorting': False,
				'searching': False,
				'width': 80,
				'type': 'label'
			},
			{
				'name': 'Total donated amount of planets',
				'index': 'total_donations',
				'sorting': True,
				'searching': False,
				'width': 60,
				'type': 'label'
			},
		]



import asyncio
import datetime
import logging

from peewee import DoesNotExist

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.conf import settings
from pyplanet.contrib import CoreContrib
from pyplanet.contrib.player.exceptions import PlayerNotFound
from pyplanet.contrib.setting.core_settings import performance_mode
from pyplanet.core.exceptions import ImproperlyConfigured
from pyplanet.core.signals import pyplanet_performance_mode_begin, pyplanet_performance_mode_end
from pyplanet.core.storage.exceptions import StorageException
from pyplanet.utils.zone import parse_path

logger = logging.getLogger(__name__)


class PlayerManager(CoreContrib):
	"""
	Player Manager.

	You can access this class in your app with:

	.. code-block:: python

		self.instance.player_manager

	With the manager you can get several useful information about the players on the server. See all the properties and methods
	below for more information.

	.. warning::

		Don't initiate this class yourself.
	"""
	def __init__(self, instance):
		"""
		Initiate, should only be done from the core instance.

		:param instance: Instance.
		:type instance: pyplanet.core.instance.Instance
		"""
		self._instance = instance
		self._performance_mode = False
		# self.lock = asyncio.Lock()

		# Online contains all currently online players.
		self._online = set()
		self._online_logins = set()

		# Counters.
		self._counter_lock = asyncio.Lock()
		self._total_count = 0
		self._players_count = 0
		self._spectators_count = 0

	@property
	def performance_mode(self):
		return self._performance_mode

	@performance_mode.setter
	def performance_mode(self, new_value):
		if self._performance_mode != new_value:
			if new_value:
				asyncio.ensure_future(
					pyplanet_performance_mode_begin.send_robust(source=dict(
						old_value=self._performance_mode, new_value=new_value
					))
				)
			else:
				asyncio.ensure_future(
					pyplanet_performance_mode_end.send_robust(source=dict(
						old_value=self._performance_mode, new_value=new_value
					))
				)
		self._performance_mode = new_value

	async def on_start(self):
		"""
		Handle startup, just before the apps will start. We will throw connects for the players so we know that the
		current playing players are also initiated correctly!
		"""
		player_list = await self._instance.gbx('GetPlayerList', -1, 0)
		await asyncio.gather(*[self.handle_connect(player['Login']) for player in player_list])

		# Load and activate blacklist.
		try:
			await self.load_blacklist()
		except:
			pass  # Ignore any exception thrown

		self._instance.signals.listen('maniaplanet:loading_map_end', self.map_loaded)

	async def map_loaded(self, *args, **kwargs):
		"""
		Reindex the current number of players and spectators.

		:param args:
		:param kwargs:
		:return:
		"""
		# Update player and spectator counters.
		player_list = await self._instance.gbx('GetPlayerList', -1, 0)

		total = 0
		specs = 0
		players = 0

		for player in player_list:
			if self._instance.game.server_is_dedicated and self._instance.game.server_player_login == player['Login']:
				continue

			try:
				info = await self._instance.gbx('GetDetailedPlayerInfo', player['Login'])
			except:
				# Player has left during this time.
				continue

			total += 1
			if info['IsSpectator']:
				specs += 1
			else:
				players += 1

		async with self._counter_lock:
			self._total_count = total
			self._spectators_count = specs
			self._players_count = players

	async def handle_connect(self, login):
		"""
		Handle a connection of a player, this call is being called inside of the Glue of the callbacks.

		:param login: Login, received from dedicated.
		:return: Database Player instance.
		:rtype: pyplanet.apps.core.maniaplanet.models.Player
		"""
		# Ignore if it's the server itself.
		if self._instance.game.server_is_dedicated and self._instance.game.server_player_login == login:
			return

		try:
			info = await self._instance.gbx('GetDetailedPlayerInfo', login)
		except:
			# Most likely too late, did disconnect directly after connecting..
			# See #126
			return
		ip, _, port = info['IPAddress'].rpartition(':')
		is_owner = login in settings.OWNERS[self._instance.process_name]

		try:
			player = await Player.get_by_login(login)
			player.last_ip = ip
			player.last_seen = datetime.datetime.now()
			player.nickname = info['NickName']
			if is_owner:
				player.level = Player.LEVEL_MASTER
			await player.save()
		except DoesNotExist:
			# Get details of player from dedicated.
			player = await Player.create(
				login=login,
				nickname=info['NickName'],
				last_ip=ip,
				last_seen=datetime.datetime.now(),
				level=Player.LEVEL_MASTER if is_owner else Player.LEVEL_PLAYER,
			)

		# Set the join time.
		player.flow.joined_at = datetime.datetime.now()

		# Update counter and state.
		async with self._counter_lock:
			player.flow.player_id = info['PlayerId']
			player.flow.team_id = info['TeamId']
			player.flow.is_spectator = bool(info['IsSpectator'])
			player.flow.is_player = not bool(info['IsSpectator'])
			player.flow.zone = parse_path(info['Path'])

			self._total_count += 1
			if player.flow.is_spectator:
				self._spectators_count += 1
			else:
				self._players_count += 1

		self._online.add(player)
		self._online_logins.add(login)
		self.performance_mode = len(self._online) >= await performance_mode.get_value()

		return player

	async def handle_info_change(self, player, is_spectator, is_temp_spectator, is_pure_spectator, target, team_id, **kwargs):
		if not player:
			return

		if player not in self._online:
			return

		async with self._counter_lock:
			if player.flow.is_spectator is True and not is_spectator:
				self._spectators_count -= 1
				self._players_count += 1

				await self._instance.signals.get_signal('maniaplanet:player_enter_player_slot').send_robust(dict(
					player=player,
				), raw=True)
			elif player.flow.is_player is True and is_spectator:
				self._spectators_count += 1
				self._players_count -= 1

				await self._instance.signals.get_signal('maniaplanet:player_enter_spectator_slot').send_robust(dict(
					player=player,
				), raw=True)

			# This is in case of desync happens. Not nice to fix, but currently one of the only options.
			if self._players_count < 0:
				self._players_count = 0
			if self._spectators_count < 0:
				self._spectators_count = 0
			if self._total_count < 0:
				self._total_count = 0

		# Update flow state.
		payload = kwargs.copy()
		payload.update(dict(
			is_spectator=is_spectator, is_temp_spectator=is_temp_spectator, is_pure_spectator=is_pure_spectator,
			target=target, team_id=team_id
		))
		player.flow.update_state(**payload)

	async def handle_disconnect(self, login):
		"""
		Handle a disconnection of a player, this call is being called inside of the Glue of the callbacks.

		:param login: Login, received from dedicated.
		:return: Database Player instance.
		:rtype: pyplanet.apps.core.maniaplanet.models.Player
		"""
		try:
			player = await Player.get_by_login(login=login)
		except:
			return

		# Update counters.
		async with self._counter_lock:
			self._total_count -= 1
			if player.flow.is_player:
				self._players_count -= 1
			else:
				self._spectators_count -= 1

		if player in self._online:
			self._online.remove(player)
		if login in self._online_logins:
			self._online_logins.remove(login)

		# Calculate the number of seconds on the server and update the total time on server.
		if player.flow.joined_at:
			time_on_server = datetime.datetime.now() - player.flow.joined_at
			player.total_playtime += int(time_on_server.total_seconds())

		try:
			del Player.CACHE[login]
		except:
			pass
		player.last_seen = datetime.datetime.now()
		await player.save()

		# Clear player/spec state.
		player.flow.reset_state()

		# Update performance mode status.
		self.performance_mode = self._total_count >= await performance_mode.get_value()

		return player

	async def get_player(self, login=None, pk=None, lock=True):
		"""
		Get player by login or primary key.

		:param login: Login.
		:param pk: Primary Key identifier.
		:param lock: Lock for a sec when receiving.
		:return: Player or exception if not found
		:rtype: pyplanet.apps.core.maniaplanet.models.Player
		"""
		try:
			if login:
				return await Player.get_by_login(login)
			elif pk:
				return await Player.get(pk=pk)
			else:
				raise PlayerNotFound('Player not found.')
		except DoesNotExist:
			if lock:
				await asyncio.sleep(4)
				return await self.get_player(login=login, pk=pk, lock=False)
			else:
				raise PlayerNotFound('Player not found.')

	async def get_player_by_id(self, identifier):
		"""
		Get player object by ID.

		:param identifier: Identifier.
		:return: Player object or None
		"""
		for player in self._online:
			if player.flow.player_id == identifier:
				return player
		return None

	async def save_blacklist(self, filename=None):
		"""
		Save the current blacklisted players to file given or fetch from config.

		:param filename: Give the filename of the blacklist, Leave empty to use the current loaded and configured one.
		:type filename: str
		:raise: pyplanet.core.exceptions.ImproperlyConfigured
		:raise: pyplanet.core.storage.exceptions.StorageException
		"""
		setting = settings.BLACKLIST_FILE
		if isinstance(setting, dict) and self._instance.process_name in setting:
			setting = setting[self._instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not filename and not setting:
			raise ImproperlyConfigured(
				'The setting \'BLACKLIST_FILE\' is not configured for this server! We can\'t save the Blacklist!'
			)
		if not filename:
			filename = setting.format(server_login=self._instance.game.server_player_login)
		
		try:
			await self._instance.gbx('SaveBlackList', filename)
		except Exception as e:
			logging.exception(e)
			raise StorageException('Can\'t save blacklist file to \'{}\'!'.format(filename)) from e

	async def save_guestlist(self, filename=None):
		"""
		Save the current guestlisted players to file given or fetch from config.

		:param filename: Give the filename of the guestlist, Leave empty to use the current loaded and configured one.
		:type filename: str
		:raise: pyplanet.core.exceptions.ImproperlyConfigured
		:raise: pyplanet.core.storage.exceptions.StorageException
		"""
		setting = settings.GUESTLIST_FILE
		if isinstance(setting, dict) and self._instance.process_name in setting:
			setting = setting[self._instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not filename and not setting:
			raise ImproperlyConfigured(
				'The setting \'GUESTLIST_FILE\' is not configured for this server! We can\'t save the Guestlist!'
			)
		if not filename:
			filename = setting.format(server_login=self._instance.game.server_player_login)

		try:
			await self._instance.gbx('SaveGuestList', filename)
		except Exception as e:
			logging.exception(e)
			raise StorageException('Can\'t save guestlist file to \'{}\'!'.format(filename)) from e


	async def load_guestlist(self, filename=None):
		"""
		Load guestlist file.

		:param filename: File to load or will get from settings.
		:raise: pyplanet.core.exceptions.ImproperlyConfigured
		:raise: pyplanet.core.storage.exceptions.StorageException
		:return: Boolean if loaded.
		"""
		setting = settings.GUESTLIST_FILE
		if isinstance(setting, dict) and self._instance.process_name in setting:
			setting = setting[self._instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not filename and not setting:
			raise ImproperlyConfigured(
				'The setting \'GUESTLIST_FILE\' is not configured for this server! We can\'t load the Guestlist!'
			)
		if not filename:
			filename = setting.format(server_login=self._instance.game.server_player_login)

		try:
			self._instance.gbx('LoadGuestList', filename)
		except Exception as e:
			logging.exception(e)
			raise StorageException('Can\'t load guestlist according the dedicated server, tried loading from \'{}\'!'.format(
				filename
			)) from e

	async def load_blacklist(self, filename=None):
		"""
		Load blacklist file.

		:param filename: File to load or will get from settings.
		:raise: pyplanet.core.exceptions.ImproperlyConfigured
		:raise: pyplanet.core.storage.exceptions.StorageException
		:return: Boolean if loaded.
		"""
		setting = settings.BLACKLIST_FILE
		if isinstance(setting, dict) and self._instance.process_name in setting:
			setting = setting[self._instance.process_name]
		if not isinstance(setting, str):
			setting = None

		if not filename and not setting:
			raise ImproperlyConfigured(
				'The setting \'BLACKLIST_FILE\' is not configured for this server! We can\'t load the Blacklist!'
			)
		if not filename:
			filename = setting.format(server_login=self._instance.game.server_player_login)

		try:
			self._instance.gbx('LoadBlackList', filename)
		except Exception as e:
			logging.exception(e)
			raise StorageException('Can\'t load blacklist according the dedicated server, tried loading from \'{}\'!'.format(
				filename
			)) from e

	@property
	def online(self):
		"""
		Online player list.
		"""
		return self._online.copy()

	@property
	def online_logins(self):
		"""
		Online player logins list.
		"""
		return self._online_logins.copy()

	@property
	def count_all(self):
		"""
		Get all player counts (players + spectators).
		"""
		return self._total_count

	@property
	def count_players(self):
		"""
		Get number of playing players.
		"""
		return self._players_count

	@property
	def count_spectators(self):
		"""
		Get number of spectating players.
		"""
		return self._spectators_count

	@property
	def max_players(self):
		"""
		Get maximum number of players.
		"""
		return self._instance.game.server_max_players

	@property
	def max_spectators(self):
		"""
		Get maximum number of spectators.
		"""
		return self._instance.game.server_max_specs



import asyncio
import logging
import math

from pyplanet.apps.config import AppConfig
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

from pyplanet.apps.contrib.transactions.view import DonationToolbarWidget

logger = logging.getLogger(__name__)


class Transactions(AppConfig):
	name = 'pyplanet.apps.contrib.transactions'
	game_dependencies = ['trackmania', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.current_bills = dict()
		self.min_donation = 10
		self.public_appreciation = 100
		self.lock = asyncio.Lock()

		self.widget = DonationToolbarWidget(self)

		self.setting_display_widget = Setting(
			'donation_widget', 'Display Donation Widget', Setting.CAT_DESIGN, type=bool, default=True,
			description='Display the donation widget to the players',
			change_target=self.reload_settings,
		)
		self.setting_display_widget_podium_only = Setting(
			'donation_widget_podium', 'Display Donation Widget only during podiums', Setting.CAT_BEHAVIOUR, type=bool,
			default=True, description='Enabling this will only display the donation widget on the podium '
									  '(requires the other setting to be enabled as well)',
			change_target=self.reload_settings
		)

	async def on_start(self):
		await self.context.setting.register(
			self.setting_display_widget, self.setting_display_widget_podium_only
		)

		await self.instance.permission_manager.register('pay', 'Pay planets to players', app=self, min_level=3)
		await self.instance.permission_manager.register('planets', 'Display amount of planets', app=self, min_level=3)

		await self.instance.command_manager.register(
			Command(command='planets', target=self.display_planets, perms='transactions:planets', admin=True),
			Command(command='pay', target=self.pay_to_player, perms='transactions:pay', admin=True)
				.add_param(name='login', required=True).add_param(name='amount', required=True),
			Command(command='donate', target=self.donate).add_param(name='amount', required=True),
		)

		# Register callback.
		self.context.signals.listen(mp_signals.other.bill_updated, self.bill_updated)

		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.flow.podium_start, self.podium_start)
		self.context.signals.listen(mp_signals.flow.podium_end, self.podium_end)
		self.context.signals.listen(mp_signals.map.map_start, self.map_start)

		# Show widget.
		if await self.setting_display_widget.get_value() and not await self.setting_display_widget_podium_only.get_value():
			await self.widget.display()

	async def reload_settings(self, *args, **kwargs):
		if await self.setting_display_widget.get_value() \
			and not await self.setting_display_widget_podium_only.get_value():
			await self.widget.display()
		elif not await self.setting_display_widget.get_value() \
			or await self.setting_display_widget_podium_only.get_value():
			await self.widget.hide()

	async def player_connect(self, player, *args, **kwargs):
		if not await self.setting_display_widget.get_value():
			return
		if await self.setting_display_widget_podium_only.get_value():
			return

		await self.widget.display(player_logins=[player.login])

	async def podium_start(self, *args, **kwargs):
		if await self.setting_display_widget.get_value() \
			and await self.setting_display_widget_podium_only.get_value():
			await self.widget.display()

	async def podium_end(self, *args, **kwargs):
		if await self.setting_display_widget.get_value() \
			and await self.setting_display_widget_podium_only.get_value():
			await self.widget.hide()

	async def map_start(self, *args, **kwargs):
		if await self.setting_display_widget.get_value() \
			and await self.setting_display_widget_podium_only.get_value():
			await self.widget.hide()

	async def display_planets(self, player, data, **kwargs):
		planets = await self.instance.gbx('GetServerPlanets')
		message = '$ff0Current server balance: $fff{}$ff0 planets.'.format(planets)
		await self.instance.chat(message, player)

	async def donate(self, player, data, **kwargs):
		try:
			async with self.lock:
				amount = abs(int(data.amount))
				if amount >= self.min_donation:
					bill_id = await self.instance.gbx('SendBill', player.login, amount, 'Donating {} planets to our server!'.format(amount), '')
					self.current_bills[bill_id] = dict(bill=bill_id, player=player, amount=amount)
				else:
					message = '$i$f00You need to donate at least $fff{}$f00 planets.'.format(self.min_donation)
					await self.instance.chat(message, player)
		except ValueError:
			message = '$i$f00The amount should be a numeric value.'
			await self.instance.chat(message, player)

	async def pay_to_player(self, player, data, **kwargs):
		try:
			amount = abs(int(data.amount))

			planets = await self.instance.gbx('GetServerPlanets')
			if amount <= (planets - 2 - math.floor(amount * 0.05)):
				async with self.lock:
					bill_id = await self.instance.gbx('Pay', data.login, amount, 'Payment from the server')
					self.current_bills[bill_id] = dict(bill=bill_id, admin=player, player=data.login, amount=-amount)
			else:
				message = '$i$f00Insufficient balance for paying $fff{}$f00 ($fff{}$f00 inc. tax) planets, only got $fff{}$f00.'.format(
					amount, (amount + 2 + math.floor(amount * 0.05)), planets
				)
				await self.instance.chat(message, player)
		except ValueError:
			message = '$z$s$fff» $i$f00The amount should be a numeric value.'
			await self.instance.chat(message, player)

	async def bill_updated(self, bill_id, state, state_name, transaction_id, **kwargs):
		async with self.lock:
			if bill_id not in self.current_bills:
				logger.debug('BillUpdated for unknown BillId {}: "{}" ({}) (TxId {})'.format(bill_id, state_name, state, transaction_id))
				return

			current_bill = self.current_bills[bill_id]
			logger.debug('BillUpdated for BillId {}: "{}" ({}) (TxId {})'.format(bill_id, state_name, state, transaction_id))
			if state == 4:
				if current_bill['amount'] > 0:
					if current_bill['amount'] > self.public_appreciation:
						message = '$f0fWe received a donation of $fff{}$f0f planets from $fff{}$z$s$f0f. Thank You!'.format(current_bill['amount'], current_bill['player'].nickname)
						await self.instance.chat(message)
					else:
						message = '$f0fYou made a donation of $fff{}$f0f planets. Thank You!'.format(current_bill['amount'])
						await self.instance.chat(message, current_bill['player'].login)

					# Update the player's total donations statistics.
					player = current_bill['player']
					player.total_donations += current_bill['amount']
					await player.save()
				else:
					message = '$f0fPayment of $fff{}$f0f planets to $fff{}$f0f confirmed!'.format(-current_bill['amount'], current_bill['player'])
					await self.instance.chat(message, current_bill['admin'].login)

				del self.current_bills[bill_id]
			elif state == 5:
				if current_bill['amount'] > 0:
					message = '$i$f00Transaction refused!'
					await self.instance.chat(message, current_bill['player'].login)
				else:
					message = '$i$f00Transaction refused!'
					await self.instance.chat(message, current_bill['admin'].login)

				del self.current_bills[bill_id]
			elif state == 6:
				if current_bill['amount'] > 0:
					message = '$i$f00Transaction failed: $fff{}$f00!'.format(state_name)
					await self.instance.chat(message, current_bill['player'].login)
				else:
					message = '$i$f00Transaction failed: $fff{}$f00!'.format(state_name)
					await self.instance.chat(message, current_bill['admin'].login)

				del self.current_bills[bill_id]



"""
MX List Views.
"""
import asyncio
import logging

from pyplanet.views.generics import ManualListView, ask_confirmation
from pyplanet.views.generics.widget import WidgetView
from pyplanet.apps.contrib.mx.exceptions import MXMapNotFound, MXInvalidResponse
from datetime import datetime
from collections import namedtuple

logger = logging.getLogger(__name__)


class MxSearchListView(ManualListView):
	def __init__(self, app, player, api):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:type app: pyplanet.apps.contrib.mx.app.RealMX
		"""
		super().__init__()
		self.manager = app.context.ui
		self.player = player
		self.app = app
		self.api = api

		self.template_name = 'mx/search.xml'
		self.response_future = asyncio.Future()

		self.cache = None
		self.has_data = False

		self.search_map = None
		self.search_author = None
		self.provide_search = True

		self.fields = [
			{
				'name': 'ID',
				'index': 'mxid',
				'sorting': True,
				'searching': False,
				'width': 15,
				'type': 'label'
			},
			{
				'name': 'Map',
				'index': 'gbxname',
				'sorting': True,
				'searching': False,
				'width': 55,
				'type': 'label'
			},
			{
				'name': 'Author',
				'index': 'author',
				'sorting': True,
				'searching': False,
				'width': 40,
				'type': 'label',
			},
			{
				'name': 'Environment',
				'index': 'envir',
				'sorting': False,
				'searching': False,
				'width': 25,
				'type': 'label',
				'safe': True
			},
			{
				'name': 'Awards',
				'index': 'awards',
				'sorting': True,
				'searching': False,
				'width': 15,
				'type': 'label'
			},
			{
				'name': 'Difficulty',
				'index': 'difficulty',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
		]

		if self.app.instance.game.game in ['tm', 'tmnext']:
			self.fields.append({
				'name': 'Length',
				'index': 'length',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			})

		self.sort_field = None
		self.child = None
		self.subscribe("mx_search", self.action_search)

	async def get_title(self):
		return '🔍 Search {}'.format(self.app.site_name)

	async def get_data(self):
		if not self.has_data:
			# First opening, request the data.
			await self.do_search(refresh=False)
			self.has_data = True

		return self.cache

	async def get_object_data(self):
		data = await super().get_object_data()
		data['search_map'] = self.search_map
		data['search_author'] = self.search_author
		return data

	async def display(self, player=None):
		return await super().display(player or self.player)

	async def get_actions(self):
		return [
			{
				'name': 'Install map',
				'type': 'label',
				'text': 'Install',
				'width': 12,
				'action': self.action_install,
				'safe': True
			}
		]

	async def action_install(self, user, values, map, *args, **kwargs):
		await self.app.instance.command_manager.execute(
			user,
			'//{} add'.format(self.app.namespace),
			str(map['mxid'])
		)

	async def action_search(self, user, action, values, *args, **kwargs):
		self.search_map = values['map']
		self.search_author = values['author']

		if values['map'] == "Search Map...":
			self.search_map = None
		if values['author'] == "Search Author...":
			self.search_author = None

		await self.do_search(self.search_map, self.search_author)

	async def do_search(self, trackname=None, authorname=None, refresh=True, **terms):
		try:
			options = {
				"api": "on",
				"mode": 0,
				"gv": 1,
				"limit": 100,
				"tpack": self.app.instance.game.dedicated_title.split("@", 1)[0]
			}

			if trackname is not None:
				options['trackname'] = trackname

			if authorname is not None:
				options['anyauthor'] = authorname

			infos = await self.api.search(options)
			if len(infos) == 0:
				raise MXMapNotFound("No results for search")

		except MXMapNotFound as e:
			message = '$f00Error requesting {}-API: Map not found!'.format(self.app.site_short_name)
			await self.app.instance.chat(message, self.player)
			logger.debug('{}-API: Map not found: {}'.format(self.app.site_short_name, str(e)))
			return None
		except MXInvalidResponse as e:
			message = '$f00Error requesting {}-API: Got an invalid response!'.format(self.app.site_short_name)
			await self.app.instance.chat(message, self.player)
			logger.warning('{}-API: Invalid response: {}'.format(self.app.site_short_name, str(e)))
			return None
		if self.app.instance.game.game in ['tm', 'tmnext']:
			self.cache = [dict(
				mxid=_map['TrackID'],
				name=_map['Name'],
				gbxname=_map['GbxMapName'] if _map['GbxMapName'] != '?' else _map['Name'],
				author=_map['Username'],
				envir=_map['EnvironmentName'],
				awards='$fff🏆 {}'.format(_map['AwardCount']) if _map['AwardCount'] > 0 else "",
				length=_map['LengthName'],
				difficulty=_map['DifficultyName'],
				maptype=_map['MapType'],
				style=_map['StyleName']
			) for _map in infos]
		else:
			self.cache = [dict(
				mxid=_map['TrackID'],
				name=_map['Name'],
				gbxname=_map['GbxMapName'] if _map['GbxMapName'] != '?' else _map['Name'],
				author=_map['Username'],
				envir=_map['EnvironmentName'],
				awards='$fff🏆 {}'.format(_map['AwardCount']) if _map['AwardCount'] > 0 else "",
				difficulty=_map['DifficultyName'],
				maptype=_map['MapType'],
				style=_map['StyleName']
			) for _map in infos]

		if refresh:
			await self.refresh(self.player)


class MxPacksListView(ManualListView):
	def __init__(self, app, player, api):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:type app: pyplanet.apps.contrib.mx.app.RealMX
		"""
		super().__init__()
		self.manager = app.context.ui
		self.player = player
		self.app = app
		self.api = api

		self.template_name = 'mx/search.xml'
		self.response_future = asyncio.Future()

		self.cache = None
		self.has_data = False

		self.search_map = None
		self.search_author = None
		self.provide_search = True

		self.fields = [
			{
				'name': 'ID',
				'index': 'mxid',
				'sorting': True,
				'searching': False,
				'width': 15,
				'type': 'label'
			},
			{
				'name': 'Pack Name',
				'index': 'name',
				'sorting': True,
				'searching': False,
				'width': 55,
				'type': 'label'
			},
			{
				'name': 'Author',
				'index': 'author',
				'sorting': True,
				'searching': False,
				'width': 40,
				'type': 'label',
			},
			{
				'name': 'Video',
				'index': 'videourl',
				'sorting': False,
				'searching': False,
				'width': 25,
				'type': 'label',
				'safe': True
			},
			{
				'name': 'Maps',
				'index': 'mapcount',
				'sorting': True,
				'searching': False,
				'width': 15,
				'type': 'label'
			}
		]

		self.sort_field = None
		self.child = None
		self.subscribe("mx_search", self.action_search)

	async def get_title(self):
		return '🔍 Search {}'.format(self.app.site_name)

	async def get_data(self):
		if not self.has_data:
			# First opening, request the data.
			await self.do_search(refresh=False)
			self.has_data = True

		return self.cache

	async def get_object_data(self):
		data = await super().get_object_data()
		data['search_map'] = self.search_map
		data['search_author'] = self.search_author
		return data

	async def display(self, player=None):
		return await super().display(player or self.player)

	async def get_actions(self):
		return [
			{
				'name': 'Install pack',
				'type': 'label',
				'text': 'Install',
				'width': 12,
				'action': self.action_install,
				'safe': True
			}
		]

	async def action_install(self, user, values, map, *args, **kwargs):
		await self.app.instance.command_manager.execute(
			user,
			'//{}pack add'.format(self.app.namespace),
			str(map['mxid'])
		)

	async def action_search(self, user, action, values, *args, **kwargs):
		self.search_map = values['map']
		self.search_author = values['author']

		if values['map'] == "Search Map...":
			self.search_map = None
		if values['author'] == "Search Author...":
			self.search_author = None

		await self.do_search(self.search_map, self.search_author)

	async def do_search(self, name=None, username=None, refresh=True, **terms):
		try:
			options = {
				"api": "on",
				"mode": 0,
				"limit": 100,
			}

			if name is not None:
				options['name'] = name

			if username is not None:
				options['username'] = username

			infos = await self.api.search_pack(options)
			if len(infos) == 0:
				raise MXMapNotFound("No results for search")

		except MXMapNotFound as e:
			message = '$f00Error requesting {}-API: Map not found!'.format(self.app.site_short_name)
			await self.app.instance.chat(message, self.player)
			logger.debug('{}-API: Map not found: {}'.format(self.app.site_short_name, str(e)))
			return None
		except MXInvalidResponse as e:
			message = '$f00Error requesting {}-API: Got an invalid response!'.format(self.app.site_short_name)
			await self.app.instance.chat(message, self.player)
			logger.warning('{}-API: Invalid response: {}'.format(self.app.site_short_name, str(e)))
			return None

		self.cache = [dict(
				mxid=_map['ID'],
				name=_map['Name'],
				author=_map['Username'],
				mapcount=_map['TrackCount'],
				typename=_map['TypeName'],
				style=_map['StyleName'],
				videourl="$l[{video}]Video$l".format(video=_map['VideoURL']) if len(_map['VideoURL']) > 0 else "",
				unreleased='{}'.format(_map['Unreleased']),
				request='{}'.format(_map['Request'])
			) for _map in infos]

		if refresh:
			await self.refresh(self.player)


class MxStatusListView(ManualListView):
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Browse'

	def __init__(self, app, api):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:type app: pyplanet.apps.contrib.mx.app.RealMX
		"""
		super().__init__()
		self.manager = app.context.ui
		self.cache = None
		self.app = app
		self.api = api

	async def get_title(self):
		return 'Server maps status on {}'.format(self.app.site_name)

	async def get_fields(self):
		fields = [
			{
				'name': 'ID',
				'index': 'index',
				'sorting': True,
				'searching': True,
				'width': 15,
				'type': 'label'
			},
			{
				'name': 'Map',
				'index': 'map_name',
				'sorting': True,
				'searching': True,
				'width': 91.5,
				'type': 'label'
			},
			{
				'name': 'On Server',
				'index': 'updated_on_server',
				'sorting': True,
				'searching': False,
				'width': 33,
				'type': 'label'
			},
			{
				'name': '{} Version'.format(self.app.site_short_name),
				'index': 'mx_version',
				'sorting': True,
				'searching': False,
				'width': 33,
				'type': 'label'
			},
			{
				'name': 'Status',
				'index': 'version_match',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
		]

		# Can only update the map via MX if it's possible to remove the current version.
		if 'admin' in self.app.instance.apps.apps:
			fields.append({
				'name': 'Update',
				'index': 'action_update_content',
				'sorting': False,
				'searching': False,
				'width': 20,
				'type': 'label',
				'action': self.action_update_map
			})

		return fields

	async def action_update_map(self, player, values, instance, **kwargs):
		# Check if the map could be updated.
		if instance['action_update'] is True:
			# Ask for confirmation.
			cancel = bool(
				await ask_confirmation(player, 'Are you sure you want to update map \'{}\'$z$s to the version from {}?'.format(
					instance['map_name'],
					self.app.site_short_name
				), size='sm')
			)
			if cancel is True:
				return

			folders = None
			if 'jukebox' in self.app.instance.apps.apps:
				# Check if the map was part of a map folder.
				# Remove the current version and add the new version to the folders.
				folders = await self.app.instance.apps.apps['jukebox'].folder_manager.get_folders_containing_map(instance['map_id'])
				if folders is not None and len(folders) != 0:
					for folder in folders:
						await self.app.instance.apps.apps['jukebox'].folder_manager.remove_map_from_folder(folder.id, instance['map_id'])

			# Remove the current version from the server.
			mock_remove = namedtuple("data", ["nr"])
			await self.app.instance.apps.apps['admin'].map.remove_map(player, mock_remove(nr=instance['map_id']))

			# Add the new version from MX.
			mock_add = namedtuple("data", ["maps"])
			added_map = await self.app.add_mx_map(player, mock_add(maps=[instance['index']]))

			# If the map could be added and was part of folders, update the folders to contain the new version.
			if added_map is not None and len(added_map) == 1:
				if folders is not None and len(folders) != 0:
					for folder in folders:
						await self.app.instance.apps.apps['jukebox'].folder_manager.add_map_to_folder(folder.id, added_map[0].id)

			# Update the current view.
			await self.refresh(player=player)

	async def get_data(self):
		if self.cache:
			print('cached')
			return self.cache

		# Determine which maps on the server could be found on MX (filter those with an ID attached).
		mx_maps_on_server = [map for map in self.app.instance.map_manager.maps if map.mx_id is not None]
		try:
			mx_maps_info = await self.api.map_info(*[map.mx_id for map in mx_maps_on_server])
		except Exception as e:
			mx_maps_info = list()
			logger.error('Could not retrieve map info from MX/TM API: {}'.format(str(e)))

		# Loop through the MX-compatible maps on the server.
		items = []
		for item in mx_maps_on_server:
			# Initialize view data fields.
			version_match = ''
			version_match_order = None
			mx_version_date = ''
			action_update = False

			# Get MX information for the map (gets None if it couldn't be found).
			mx_map = next((mx_map_info for mx_map_info in mx_maps_info if mx_map_info[0] == item.mx_id), None)

			if mx_map is None:
				version_match = 'Not on {}'.format(self.app.site_short_name)
				version_match_order = 1
			else:
				date_format = '%Y-%m-%dT%H:%M:%S'
				if '.' in mx_map[1]['UpdatedAt']:
					date_format = '%Y-%m-%dT%H:%M:%S.%f'
				mx_version_date = datetime.strptime(mx_map[1]['UpdatedAt'], date_format).strftime("%Y-%m-%d %H:%M:%S")
				mx_map_uid = mx_map[1]['TrackUID'] if 'TrackUID' in mx_map[1] else mx_map[1]['MapUID']

				if mx_map_uid == item.uid:
					version_match = '$0a0Up-to-date'
					version_match_order = 2
				else:
					version_match = '$00fNew version'
					version_match_order = 0
					action_update = True

			action_update_content = '🔁 Update' if action_update else '          -'
			items.append({'map_id': item.id, 'index': item.mx_id, 'map_name': item.name, 'version_match': version_match, 'version_match_order': version_match_order,
				'updated_on_server': item.updated_at, 'mx_version': mx_version_date, 'action_update': action_update, 'action_update_content': action_update_content})

		# Initially sort the maps based on the 'version_match_order': New version -> Not on MX -> Up-to-date.
		items.sort(key=lambda x: x['version_match_order'])
		self.cache = items
		return self.cache

	async def destroy(self):
		self.cache = None
		return await super().destroy()

	def destroy_sync(self):
		self.cache = None
		return super().destroy_sync()


class MxAwardWidget(WidgetView):
	"""
	Award widget.
	"""

	widget_x = 63
	widget_y = -67.5
	z_index = 100
	template_name = 'mx/award.xml'
	mx_id = None

	def __init__(self, app):
		"""
		Initializes the widget.

		:param app: the MX application (plugin)
		"""

		super().__init__(app.context.ui)
		self.app = app
		self.id = 'pyplanet__widgets_mxaward'

	async def get_context_data(self):
		"""
		Called to request data to display in the widget.
		"""

		context = await super().get_context_data()

		context.update({
			'map_name': self.app.instance.map_manager.current_map.name,
			'site_name': self.app.site_short_name,
			'award_url': "{}/maps/{}/#award".format(self.app.api.base_url(), self.mx_id)
		})

		return context



import asynctest
import datetime

from pyplanet.core import Controller


class TestPermissions(asynctest.TestCase):
	async def test_registration(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()

		from pyplanet.apps.core.maniaplanet.models import Player
		await instance.permission_manager.register(
			'test1',
			namespace='tst1',
			min_level=Player.LEVEL_PLAYER
		)
		assert bool(await instance.permission_manager.get_perm('tst1', 'test1'))

	async def test_checking(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		from pyplanet.apps.core.maniaplanet.models import Player
		await instance.permission_manager.register(
			'sample1',
			namespace='tst1',
			min_level=Player.LEVEL_PLAYER
		)
		await instance.permission_manager.register(
			'sample2',
			namespace='tst1',
			min_level=Player.LEVEL_OPERATOR
		)
		await instance.permission_manager.register(
			'sample3',
			namespace='tst1',
			min_level=Player.LEVEL_ADMIN
		)
		await instance.permission_manager.register(
			'sample4',
			namespace='tst1',
			min_level=Player.LEVEL_MASTER
		)

		# Try to fetch the command.
		try:
			player1 = await Player.get(login='sample-1')
		except:
			player1 = await Player.create(login='sample-1', nickname='sample-1', level=Player.LEVEL_MASTER)
		try:
			player2 = await Player.get(login='sample-2')
		except:
			player2 = await Player.create(login='sample-2', nickname='sample-2', level=Player.LEVEL_PLAYER)

		assert await instance.permission_manager.has_permission(player1, 'tst1:sample1')
		assert await instance.permission_manager.has_permission(player1, 'tst1:sample2')
		assert await instance.permission_manager.has_permission(player1, 'tst1:sample3')
		assert await instance.permission_manager.has_permission(player1, 'tst1:sample4')

		assert await instance.permission_manager.has_permission(player2, 'tst1:sample1')
		assert not await instance.permission_manager.has_permission(player2, 'tst1:sample2')
		assert not await instance.permission_manager.has_permission(player2, 'tst1:sample3')
		assert not await instance.permission_manager.has_permission(player2, 'tst1:sample4')



import asyncio
import re

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.views import TemplateView


class AlertView(TemplateView):
	"""
	The AlertView can be used to show several generic alerts to a player. You can use 3 different sizes, and adjust the
	message text.

	The 3 sizes:
	sm, md and lg.
	"""

	template_name = 'core.views/generics/alert.xml'

	SIZES = dict(
		sm={
			'top__pos': '0 17',
			'top__size': '126.5 8',
			'box__size': '120 25',
			'bottom__pos': '0 -12',
			'bottom__size': '120 2',
			'text__pos': '-47.5 6.25',
			'text__size': '100 4',
			'button_0__pos__left': -32,
			'button_0__pos__top': -4,
			'button_1__pos__left': 32,
			'button_1__pos__top': -4,
		},
		md={
			'top__pos': '0 30',
			'top__size': '156.5 8',
			'box__size': '150 50',
			'bottom__pos': '0 -24',
			'bottom__size': '150 2',
			'text__pos': '-67.5 6.25',
			'text__size': '135 24',
			'button_0__pos__left': -32,
			'button_0__pos__top': -14,
			'button_1__pos__left': 32,
			'button_1__pos__top': -14,
		},
		lg={
			'top__pos': '0 55',
			'top__size': '206.5 8',
			'box__size': '200 100',
			'bottom__pos': '0 -50',
			'bottom__size': '200 2',
			'text__pos': '-92.5 6.25',
			'text__size': '185 74',
			'button_0__pos__left': -52,
			'button_0__pos__top': -40,
			'button_1__pos__left': 52,
			'button_1__pos__top': -40,
		},
	)

	def __init__(
		self, message, size='md', buttons=None, manager=None, target=None, **data
	):
		"""
		Create an AlertView instance.

		:param message: The message to display to the end-user, Use ``\\n`` for new lines. You can use symbols from FontAwesome
						by using Unicode escaped strings.
		:param size: Size to use, this parameter should be a string, and one of the following choices:
					 'sm', 'md' or 'lg. Defaults to 'md'.
		:param buttons: Buttons to display, Should be an array with dictionary which contain: name.
		:param manager: UI Manager to use, You should always keep this undefined unless you know what your doing!
		:param target: Target coroutine method called as handle of button clicks.

		:type message: str
		:type title: str
		:type size: str
		:type buttons: list
		:type manager: pyplanet.core.ui._BaseUIManager
		"""
		from pyplanet.core import Controller

		super().__init__(manager or Controller.instance.ui_manager)
		self.disable_alt_menu = True
		sizes = self.SIZES[size]

		if not buttons:
			buttons = [{'name': 'OK'}]

		self.target = target
		self.response_future = asyncio.Future()

		self.data = dict(
			message=message,
			buttons=buttons,
			sizes=sizes,
		)
		self.data.update(data)

	async def wait_for_reaction(self):  # pragma: no cover
		"""
		Wait for reaction or input and return it.

		:return: Returns the button clicked or the input value string of the user.
		"""
		return await self.response_future

	async def handle(self, player, action, values, **kwargs):  # pragma: no cover
		await self.close(player)

		# Try to parse the button id instead of the whole action string.
		button = action
		try:
			match = re.search('button_([0-9]+)$', action)
			if len(match.groups()) == 1:
				button = match.group(1)
		except:
			pass

		if not self.response_future.done():
			self.response_future.set_result(button)

		if self.target:
			await self.target(player, action, values, **kwargs)

	async def close(self, player, **kwargs):  # pragma: no cover
		"""
		Close the alert.
		"""
		await self.hide(player_logins=[player.login])


class PromptView(AlertView):
	"""
	The PromptView is like the AlertView but can ask for a text entry.

	The 3 sizes:
	sm, md and lg.

	You can listen for the results of the players input with the ``wait_for_input()`` async handler (future).
	Example:

	.. code-block:: python

		prompt = PromptView('Please enter your name')
		await prompt.display(['login'])

		user_input = await prompt.wait_for_input()
		print(user_input)


	You can do validations before it's okay with giving a function to the argument ``validator``. Example:

	.. code-block:: python

		def my_validator(value):
			try:
				int(value)
				return True, None
			except:
				return False, 'Value should be an integer!'

		prompt = PromptView('Please enter your name', validator=my_validator)
		await prompt.display(['login'])

		user_input = await prompt.wait_for_input()
		print(user_input)

	"""

	template_name = 'core.views/generics/prompt.xml'

	SIZES = dict(
		sm={
			'top__pos': '0 24',
			'top__size': '126.5 8',
			'box__size': '120 40',
			'bottom__pos': '0 -20',
			'bottom__size': '120 2',
			'text__pos': '-47.5 11',
			'text__size': '100 4',
			'input__pos': '0 0',
			'input__size': '100 7',
			'button_0__pos__left': -32,
			'button_0__pos__top': -12,
			'button_1__pos__left': 32,
			'button_1__pos__top': -12,
		},
		md={
			'top__pos': '0 38',
			'top__size': '156.5 8',
			'box__size': '150 67',
			'bottom__pos': '0 -34',
			'bottom__size': '150 2',
			'text__pos': '-67.5 11',
			'text__size': '135 24',
			'input__pos': '0 -10',
			'input__size': '135 7',
			'button_0__pos__left': -32,
			'button_0__pos__top': -24,
			'button_1__pos__left': 32,
			'button_1__pos__top': -24,
		},
		lg={
			'top__pos': '0 67',
			'top__size': '206.5 8',
			'box__size': '200 125',
			'bottom__pos': '0 -62',
			'bottom__size': '200 2',
			'text__pos': '-92.5 11',
			'text__size': '185 74',
			'input__pos': '0 -40',
			'input__size': '185 7',
			'button_0__pos__left': -52,
			'button_0__pos__top': -52,
			'button_1__pos__left': 52,
			'button_1__pos__top': -52,
		},
	)

	def __init__(self, message, size='md', buttons=None, manager=None, default='', validator=None):
		super().__init__(message, size, buttons, manager)
		self.disable_alt_menu = True

		self.default = default
		self.validator = validator or self.validate_input

		self.data['default'] = self.default

	async def wait_for_input(self):  # pragma: no cover
		"""
		Wait for input and return it.

		:return: Returns the string value of the user.
		"""
		return await self.response_future

	def validate_input(self, value):  # pragma: no cover
		if not value or len(value) == 0:
			return False, 'Empty value given!'
		return True, None

	async def handle(self, player, action, values, **kwargs):  # pragma: no cover
		self.data['errors'] = ''
		value = self.default
		if 'prompt_value' in values:
			value = values['prompt_value']

		valid, message = self.validator(value)

		if valid:
			await self.close(player)
			if not self.response_future.done():
				self.response_future.set_result(value)
			return

		self.data['errors'] = message
		await self.display([player.login])


# Util methods.
async def ask_confirmation(player, message, size='md', buttons=None):  # pragma: no cover
	"""
	Ask the player for confirmation and return the button number (0 is first button).

	:param player: Player login or instance.
	:param message: Message to display.
	:param size: Size, could be 'sm', 'md', or 'lg'.
	:param buttons: Buttons, optional, default is yes and no.
	:return: Number of button that is clicked.
	"""
	buttons = buttons or [{'name': 'Yes'}, {'name': 'No'}]
	view = AlertView(message, size, buttons)
	if isinstance(player, Player):
		player = player.login
	await view.display(player_logins=[player])
	reaction = await view.wait_for_reaction()
	try:
		reaction = int(reaction)
	except:
		reaction = None
	del view
	return reaction


async def ask_input(player, message, size='md', buttons=None, default=None, validator=None):  # pragma: no cover
	"""
	Ask the player a question and prompt for input.

	:param player: Player login or instance.
	:param message: Message to display.
	:param size: Size, could be 'sm', 'md', or 'lg'
	:param buttons: Buttons, optional, default is ok.
	:param default: The default and pre-filled value. Default empty.
	:param validator: Validator method, default is only checking if the input isn't empty.
	:return: Input given by the user.
	"""
	buttons = buttons or [{'name': 'OK'}]
	view = PromptView(message, size, buttons, default=default, validator=validator)
	if isinstance(player, Player):
		player = player.login
	await view.display(player_logins=[player])
	output = await view.wait_for_input()
	del view
	return output


async def show_alert(player, message, size='md', buttons=None):  # pragma: no cover
	"""
	Show an alert to the player with given details. This is a shortcut method for the class itself.

	:param player: Player login or instance.
	:param message: Message in string.
	:param size: Size, could be 'sm', 'md', or 'lg'.
	:param buttons: Buttons, optional, default is 'OK'.
	:return: Number of the clicked button. (in Future).
	"""
	buttons = buttons or [{'name': 'OK'}]
	view = AlertView(message, size, buttons)
	if isinstance(player, Player):
		player = player.login
	await view.display(player_logins=[player])
	reaction = await view.wait_for_reaction()
	try:
		reaction = int(reaction)
	except:
		reaction = None
	del view
	return reaction



from pyplanet.apps.config import AppConfig


class {{ camel_case_app_name }}App(AppConfig):

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

	async def on_init(self):
		await super().on_init()

	async def on_start(self):
		await super().on_start()

	async def on_stop(self):
		await super().on_stop()

	async def on_destroy(self):
		await super().on_destroy()



"""
Toolbar interface and methods.
"""
import logging

from pyplanet.apps.contrib.admin.views.toolbar import ToolbarView
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

logger = logging.getLogger(__name__)


class ToolbarAdmin:
	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.admin.app.Admin
		"""
		self.app = app
		self.instance = app.instance

		self.view = ToolbarView(self.app)

		self.setting_enable_toolbar = Setting(
			'enable_toolbar', 'Enable the admin toolbar', Setting.CAT_DESIGN, type=bool,
			description='Enable and show the admin toolbar for admins (all players with level above player level).',
			default=True
		)

	async def on_start(self):
		await self.app.context.setting.register(self.setting_enable_toolbar)

		# Listen to connect event.
		self.app.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		# Register commands.
		await self.instance.command_manager.register(
			Command(command='toolbar', target=self.toggle_toolbar, admin=True, description='Toggles the admin toolbar.')
		)

		# Display to all current online admins.
		if await self.setting_enable_toolbar.get_value():
			for player in self.app.instance.player_manager.online:
				if player.level > player.LEVEL_PLAYER:
					await self.view.display([player.login])

	async def player_connect(self, player, is_spectator, source, **kwargs):
		if player.level > player.LEVEL_PLAYER and await self.setting_enable_toolbar.get_value():
			await self.view.display([player.login])

	async def toggle_toolbar(self, player, *args, **kwargs):
		if player.level <= player.LEVEL_PLAYER:
			return

		if player.login in self.view._is_player_shown:
			await self.view.hide([player.login])
		else:
			await self.view.display([player.login])



from pyplanet.apps.config import AppConfig
from pyplanet.contrib.command import Command

from pyplanet.apps.contrib.players.views import PlayerListView
from pyplanet.apps.core.maniaplanet.callbacks import player as player_signals
from pyplanet.contrib.player.exceptions import PlayerNotFound
from pyplanet.contrib.setting import Setting


class Players(AppConfig):
	name = 'pyplanet.apps.contrib.players'
	game_dependencies = ['trackmania', 'trackmania_next', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.setting_enable_join_msg = Setting(
			'enable_join_message', 'Enable player join messages', Setting.CAT_FEATURES, type=bool, default=True,
			description='Enable this to announce that a player has joined the server.'
		)
		self.setting_enable_leave_msg = Setting(
			'enable_leave_message', 'Enable player leave messages', Setting.CAT_FEATURES, type=bool, default=True,
			description='Enable this to announce that a player has left the server.'
		)

	async def on_start(self):
		await self.instance.command_manager.register(
			Command(command='players', target=self.player_list, description='Displays the list of players currently online.'),
			Command(command='laston', aliases=['lastseen'], target=self.command_laston,
					description='Displays the last time the provided player was online on this server.').add_param(name='login', required=True),
		)
		await self.context.setting.register(
			self.setting_enable_join_msg, self.setting_enable_leave_msg
		)

		player_signals.player_connect.register(self.player_connect)
		player_signals.player_disconnect.register(self.player_disconnect)

	async def player_list(self, player, data, **kwargs):
		view = PlayerListView(self)
		await view.display(player=player.login)

	async def command_laston(self, player, data, **kwargs):
		try:
			laston_player = await self.instance.player_manager.get_player(data.login, None, False)
			if laston_player.last_seen is not None:
				message = '$ff0Player $fff{}$z$s$ff0 was last online on $fff{}$ff0 at $fff{}$ff0.'.format(laston_player.nickname, laston_player.last_seen.strftime('%d-%m-%Y'), laston_player.last_seen.strftime('%H:%M:%S'))
				await self.instance.chat(message, player)
			else:
				message = '$ff0Player $fff{}$z$s$ff0 has not been seen online (yet).'.format(laston_player.nickname)
				await self.instance.chat(message, player)
		except PlayerNotFound:
			message = '$i$f00Unknown login!'
			await self.instance.chat(message, player)

	async def player_connect(self, player, **kwargs):
		if not await self.setting_enable_join_msg.get_value():
			return

		if player.flow.is_spectator:
			message = '$ff0{} $fff{}$z$s$ff0 joined the server as spectator! {}'
		else:
			message = '$ff0{} $fff{}$z$s$ff0 joined the server! {}'
		await self.instance.chat(
			message.format(
				player.get_level_string(), player.nickname,
				'Nation: $fff{}'.format(player.flow.zone.country) if player.flow.zone else ''
			)
		)

	async def player_disconnect(self, player, **kwargs):
		if not await self.setting_enable_leave_msg.get_value():
			return

		await self.instance.chat(
			'$ff0{} $fff{}$z$s$ff0 left the server! {}'.format(
				player.get_level_string(), player.nickname,
				'Nation: $fff{}'.format(player.flow.zone.country) if player.flow.zone else ''
			)
		)



from pyplanet.views import TemplateView


class ControllerView(TemplateView):
	template_name = 'core.pyplanet/controller.xml'

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.id = 'pyplanet__controller'

		self.subscribe('f8', self.action_f8)

	async def get_context_data(self):
		from pyplanet.core import Controller
		context = await super().get_context_data()
		context['game'] = Controller.instance.game.game

		context['chat_pos'] = '-160.25 -63.75'
		if context['game'] != 'tm':
			context['chat_pos'] = '-160.25 -39.75'

		return context

	async def display(self, **kwargs):
		return await super().display(**kwargs)

	async def action_f8(self, player, *args, **kwargs):
		await self.manager.instance.chat(
			'$ff0Toggling visibility. You can show/hide with F8, and show/hide when driving with F9', player
		)



import os
import asynctest

from pyplanet.core import Controller


class TestStorageManager(asynctest.TestCase):

	def __init__(self, *args, **kwargs):
		from pyplanet.conf import settings
		self.tmp_dir = settings.TMP_PATH
		self.tmp_file = os.path.join(self.tmp_dir, 'test-{}'.format(id(self)))
		self.instance = Controller.prepare(name='default').instance
		super().__init__(*args, **kwargs)

	async def test_init(self):
		assert self.instance.storage.driver

	async def test_exists(self):
		assert await self.instance.storage.driver.exists(__file__) is True

	async def test_touch(self):
		file = '{}-1.txt'.format(self.tmp_file)
		await self.instance.storage.driver.touch(file)
		assert await self.instance.storage.driver.exists(file) is True
		await self.instance.storage.driver.remove(file)
		assert await self.instance.storage.driver.exists(file) is False

	async def test_remove(self):
		file = '{}-2.txt'.format(self.tmp_file)
		await self.instance.storage.driver.touch(file)
		assert await self.instance.storage.driver.exists(file) is True
		await self.instance.storage.driver.remove(file)
		assert await self.instance.storage.driver.exists(file) is False

	async def test_open_read_write(self):
		file = '{}-3.txt'.format(self.tmp_file)
		await self.instance.storage.driver.touch(file)

		async with self.instance.storage.driver.open(file, 'w') as fh:
			await fh.write('Test OK')

		async with self.instance.storage.driver.open(file) as fh:
			assert await fh.read() == 'Test OK'



import asyncio
import logging
import platform

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.pyplanet.dev import DevComponent
from pyplanet.apps.core.pyplanet.setting import SettingComponent
from pyplanet.apps.core.pyplanet.toolbar import ToolbarComponent
from pyplanet.apps.core.pyplanet.views.command import CommandsListView
from pyplanet.apps.core.pyplanet.views.controller import ControllerView
from pyplanet.conf import settings
from pyplanet.contrib.command import Command

from pyplanet import __version__ as version
from pyplanet.utils import semver
from pyplanet.utils.functional import batch
from pyplanet.utils.pip import Pip
from pyplanet.utils.releases import UpdateChecker
from pyplanet.views.generics import ask_confirmation


class PyPlanetConfig(AppConfig):
	core = True

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		# Initiate components.
		self.setting = SettingComponent(self)
		self.dev = DevComponent(self)
		self.toolbar = ToolbarComponent(self)

		# Initiate app (global) view.
		self.controller_view = ControllerView(manager=self.context.ui)

		# Status.
		self.in_progress = False

	async def on_init(self):
		# Call components.
		await self.setting.on_init()
		await self.dev.on_init()
		await self.toolbar.on_init()

	async def on_start(self):
		# Call components.
		await self.setting.on_start()
		await self.dev.on_start()
		await self.toolbar.on_start()

		# Change some ui elements positions and visibility.
		if self.instance.game.game in ['tm', 'sm']:
			self.instance.ui_manager.properties.set_visibility('live_info', False)
			self.instance.ui_manager.properties.set_attribute('live_info', 'pos', '-125. 84. 5.')
			self.instance.ui_manager.properties.set_attribute('warmup', 'pos', '86., 87., 5.')
		else:
			self.instance.ui_manager.properties.set_visibility('Race_Checkpoint', False)
			self.instance.ui_manager.properties.set_visibility('Race_RespawnHelper', False)
			self.instance.ui_manager.properties.set_attribute('Race_BestRaceViewer', 'position', [135, -30])
			self.instance.ui_manager.properties.set_attribute('Race_Countdown', 'position', [155, -15])
			self.instance.ui_manager.properties.set_attribute('Race_LapsCounter', 'position', [155.7, -77])
			self.instance.ui_manager.properties.set_attribute('Race_LapsCounter', 'scale', 0.7)
			self.instance.ui_manager.properties.set_attribute('Race_TimeGap', 'position', [44, -45])
			self.instance.ui_manager.properties.set_attribute('Rounds_SmallScoresTable', 'position', [-160, 8])
			self.instance.ui_manager.properties.set_attribute('Race_WarmUp', 'position', [155.7, -65])
			self.instance.ui_manager.properties.set_attribute('Race_WarmUp', 'scale', 0.7)
			self.instance.ui_manager.properties.set_attribute('Race_ScoresTable', 'scale', 0.9)
			self.instance.ui_manager.properties.set_attribute('Race_SpectatorBase_Commands', 'position', [50, -87])
			await self.instance.ui_manager.properties.send_properties()

		# Display logo.
		await self.controller_view.display()

		# Listeners.
		self.context.signals.listen('maniaplanet:player_connect', self.on_connect)
		await self.instance.command_manager.register(
			Command('version', self.chat_version,
					description='Displays the current server and PyPlanet versions and the active PyPlanet plugins.'),
			Command('upgrade', self.chat_upgrade, admin=True, description='Upgrade PyPlanet installation (Experimental)')
				.add_param('to_version', type=str, default=None, required=False, help='Upgrade to specific version'),

			Command('help', target=self.chat_help, description='Shows a chat-based list with all available commands.')
				.add_param('command', nargs='*', required=False),
			Command('help', target=self.chat_help, admin=True, description='Shows a chat-based list with all available admin commands.')
				.add_param('command', nargs='*', required=False),
			Command('helpall', target=self.chat_helpall, description='Shows all available commands in a list window.'),
			Command('helpall', target=self.chat_helpall, admin=True, description='Shows all available admin commands in a list window.')
		)

	async def on_connect(self, player, **kwargs):
		await self.controller_view.display(player_logins=[player.login])

	async def chat_version(self, player, *args, **kwargs):
		message = '$ff0PyPlanet: $fff{}$ff0 (Python $fff{}$ff0), current apps: $fff'.format(version, platform.python_version())
		message += '$ff0, $fff'.join(self.instance.apps.apps.keys())
		await self.instance.chat(message, player)

	async def chat_upgrade(self, player, data, *args, **kwargs):
		package = 'pyplanet'
		to_version = data.to_version
		if player.level != 3:
			return await self.instance.chat('$f00Only MasterAdmin can upgrade installation!', player)
		if not settings.SELF_UPGRADE:
			return await self.instance.chat('$f00In-game upgrade method is disabled by the hoster!', player)

		# Getting PIP.
		pip = Pip()
		await self.instance.gbx.multicall(
			self.instance.chat('$09fPIP:$fff pip-command: {}'.format(pip.command), player),
			self.instance.chat('$09fPIP:$fff pip-status: {}'.format(
				'PIP is ready and supported' if pip.is_supported else '$f00Unsupported! Use manual method!'
			), player)
		)
		if not pip.is_supported:
			return

		# Validate new version
		if to_version:
			if to_version not in UpdateChecker.releases:
				return await self.instance.chat(
					'$f00The version \'{}\' is not known to be a PyPlanet version!'.format(to_version), player
				)
		else:
			to_version = UpdateChecker.latest

		# Check if to_version is actual newer.
		from pyplanet import __version__ as current_version
		if semver.compare(to_version, current_version) < 0:
			return await self.instance.chat(
				'$f00Downgrading is not possible with the in-game upgrade method!'.format(to_version), player
			)

		# Ask the user, give warnings.
		cancel = bool(await ask_confirmation(
			player,
			'Are you sure you want to upgrade PyPlanet from \'{}\' to \'{}\'?'
			'\n$f00$o$wWARNING:$g$n This method is experimental, you can break your installation! '
			'We advice to use the PIP method over the in-game upgrade method! Please '
			'be careful with big release updates!'.format(current_version, to_version),
			size='md',
		))

		if cancel:
			return

		# Execute the upgrade.
		await self.instance.gbx.multicall(
			self.instance.chat('$09fUpgrade$fff: PyPlanet is about to upgrade to {}'.format(to_version)),
			self.instance.chat('$09fUpgrade$fff: Please wait, installing new version with $09fpip$fff...'.format(to_version))
		)

		self.in_progress = True
		asyncio.ensure_future(self.send_updates())
		await asyncio.sleep(2)

		if settings.DEBUG:
			logging.getLogger(__name__).warning('DEBUG MODE, NOT ACTUALLY UPGRADING PYPLANET!!!!!!!!!!!!!!!!!!!')
			package = 'example'
			to_version = None

		statuscode, stdout, stderr = pip.install(package, target_version=to_version)

		self.in_progress = False
		await self.instance.chat('.', raw=True)

		if statuscode != 0:
			return await self.instance.gbx.multicall(
				self.instance.chat(
					'$09fUpgrade$fff: Upgrade failed! Please look at the logfiles or the following lines for information'
				),
				self.instance.chat(
					'$09fUpgrade$fff: PIP (truncated): $n{}'.format(stderr.decode()[-100:])
				),
				self.instance.chat(
					'$09fUpgrade$fff: Please use the manual PIP method for upgrading and check if your installation isn\'t broken'
				)
			)

		# Send successful message.
		await self.instance.chat(
			'$09fUpgrade$fff: PIP installation of PyPlanet succeeded! Please wait...'
		)

		# Reboot PyPlanet
		await self.instance.chat(
			'$09fUpgrade$fff: Restarting PyPlanet...'
		),
		exit(50)

	async def send_updates(self):
		await self.instance.chat(
			'$09fUpgrade$fff: $'
		)
		while self.in_progress:
			await self.instance.chat(
				'.$',
				raw=True
			)
			await asyncio.sleep(0.05)

	async def chat_helpall(self, player, data, raw, command):
		return

	async def chat_help(self, player, data, raw, command):
		# Show usage of a single command, given as second or more params.
		if data.command:
			is_admin = bool(command.admin)
			cmd_args = data.command

			# HACK: Add / before an admin command to match the right command.
			if is_admin and cmd_args:
				cmd_args[0] = '/{}'.format(cmd_args[0])

			cmd = await self.instance.command_manager.get_command_by_command_text(cmd_args)
			if cmd:
				await self.instance.chat(
					'$z$s{}'.format(cmd.usage_text),
					player
				)
			else:
				await self.instance.chat('Command help not found', player)
			return

		# All commands.
		commands = await self.instance.command_manager.help_entries(player, command.admin)

		# Prepare and send the calls.
		calls = list()
		for cmds in batch(commands, 7):
			help_texts = [str(c) for c in cmds]
			calls.append(
				self.instance.chat(
					'$z$s{}'.format(' | '.join(help_texts)),
					player.login
				)
			)

		await self.instance.gbx.multicall(
			self.instance.chat(
				'$z$sCommand list. Help per command: /{}help [command]'.format('/' if command.admin else ''),
				player.login
			),
			*calls
		)

	async def chat_helpall(self, player, data, raw, command):
		window = CommandsListView(self, player, command.admin)
		await window.display(player=player)
		return



import json
from asyncio import iscoroutinefunction

from pyplanet.apps.core.pyplanet.models.setting import Setting as SettingModel
from pyplanet.contrib.setting.exceptions import TypeUnknownException, SerializationException, SettingException


class Setting:
	"""
	The setting class is for defining a setting for the end-user.
	This setting can be changed with /settings and //settings.

	With this class you can define or manage your setting that is going to be public for all other apps and end-user.

	You can get notified of changes with the ``change_target`` in the init of this class. Point this to a method (async or sync)
	with the following params: ``old_value`` and ``new_value``.

	Example:

	.. code-block:: python

		my_setting = Setting(
			'dedimania_code', 'Dedimania Server Code', Setting.CAT_KEYS, type=str,
			description='The secret dedimania code. Get one at $lhttp://dedimania.net/tm2stats/?do=register',
			default=None
		)

		my_other_setting = Setting(
			'sample_boolean', 'Booleans for the win!', Setting.CAT_BEHAVIOUR, type=bool, description='Example',
		)

	"""

	CAT_GENERAL = 'General'
	CAT_KEYS = 'Keys'
	CAT_DESIGN = 'Design'
	CAT_POSITION = 'Position'
	CAT_BEHAVIOUR = 'Behaviour'
	CAT_FEATURES = 'Features'
	CAT_OTHER = 'Other'
	ALL_CATEGORIES = [CAT_GENERAL, CAT_DESIGN, CAT_POSITION, CAT_BEHAVIOUR, CAT_FEATURES, CAT_KEYS, CAT_OTHER]

	def __init__(
		self, key: str, name: str, category: str, type=str, description: str = None, choices=None, default=None,
		change_target=None
	):
		"""
		Create setting with properties.

		:param key: Key of setting, this is mainly only used for the backend and for referencing the setting.
					You should keep this unique in your app!
		:param name: Name of the setting that will be displayed as a small label to the player.
		:param category: Category from Categories.*. Must be provided!
		:param type: Type of value to expect, use python types here. str by default.
		:param description: Description to provide help and instructions to the player.
		:param choices: List or tuple with choices, only when wanting to restrict values to selected options.
		:param default: Default value if not provided from database. This will be returned. Defaults to None.
		:param change_target: Target method to call when the setting value has been changed.
		"""
		if category not in self.ALL_CATEGORIES:
			raise SettingException('Invalid category. Must be an category in the Categories static class.')
		# Prepare property for app specific setting. Will be injected by the register command.
		self.app_label = None

		self.key = key
		self.name = name
		self.description = description
		self.category = category
		self.default = default
		self.type = type
		self.choices = choices
		self.change_target = change_target

		# Prepare the model instance here. This will be filled once it's fetched for the first time (or inited).
		self._instance = None
		self._value = (False, None)

	async def initiate_setting(self):
		"""
		Initiate database record for setting.
		"""
		return await SettingModel.get_or_create_from_info(
			key=self.key, app=self.app_label, category=self.category, name=self.name, description=self.description,
			value=None
		)

	def unserialize_value(self, value):
		"""
		Unserialize the datastorage value to the python value, based on the type of the setting.

		:param value: Value from database.
		:return: Python value.
		:raise pyplanet.contrib.setting.exceptions.SerializationException: SerializationException
		"""
		if value is None:
			return self.default

		try:
			if self.type == str:
				return str(value)
			elif self.type == int:
				return int(value)
			elif self.type == float:
				return float(value)
			elif self.type == bool:
				return bool(value)
			elif self.type == list or self.type == set or self.type == dict:
				return json.loads(value)
			else:
				raise TypeUnknownException('The type \'{}\' is unknown!'.format(self.type))
		except TypeUnknownException:
			raise
		except Exception as e:
			raise SerializationException('Error with unserialization of the setting \'{}\''.format(str(self))) from e

	def serialize_value(self, value):
		"""
		Serialize the python value to the data store value, based on the type of the setting.

		:param value: Python Value.
		:return: Database Value
		"""
		# Always set to Null, so we get the default value back.
		if value is None:
			return value

		# Empty value, set the default.
		if value == '':
			return self.default

		if self.choices and value not in self.choices:
			raise SerializationException('Value given is not in the predefined choices!')

		try:
			if self.type == int:
				value = int(value)
			elif self.type == float:
				value = float(value)
			elif self.type == bool:
				if value == '1' or value == 1 or value == '0' or value == 0:
					value = bool(int(value))
		except:
			pass

		if type(value) != self.type:
			raise SerializationException(
				'Your given value is not of the type you specified! \'{}\' != \'{}\''.format(type(value), self.type)
			)

		if self.type == list or self.type == set or self.type == dict:
			return json.dumps(value)
		if self.type == bool:
			return value
		return str(value)

	@property
	def type_name(self):
		"""
		Get the name of the specified type in string format, suited for displaying to end-user.

		:return: User friendly name of type.
		"""
		if self.type == str:
			return 'string'
		elif self.type == int:
			return 'integer'
		elif self.type == float:
			return 'float'
		elif self.type == bool:
			return 'boolean'
		elif self.type == list or self.type == set:
			return 'list'
		elif self.type == dict:
			return 'dict'
		else:
			return 'unknown'

	async def get_value(self, refresh=False):
		"""
		Get the value or the default value for the setting model.

		:param refresh: Force a refresh of the value.
		:return: Value in the desired type and unserialized from database/storage.
		:raise: NotFound / SerializationException
		"""
		if not self._value[0] or refresh is True:
			model = await self.get_model()
			self._value = (True, self.unserialize_value(model.value))
		return self._value[1]

	async def set_value(self, value):
		"""
		Set the value, this will serialize and save the setting to the data storage.

		:param value: Python value input.
		:raise: NotFound / SerializationException
		"""
		old_value = self._value[0] if self._value and len(self._value) > 0 else None

		model = await self.get_model()
		model.value = self.serialize_value(value)
		self._value = (True, model.value)
		await model.save()

		# Call the change target.
		if self.change_target and callable(self.change_target):
			if iscoroutinefunction(self.change_target):
				await self.change_target(old_value, model.value)
			else:
				self.change_target(old_value, model.value)

	async def clear(self):
		"""
		Clear the value in the data storage. This will set the value to None, and will return the default value on
		request of data.

		:raise: NotFound / SerializationException
		"""
		return await self.set_value(None)

	async def get_model(self):
		"""
		Get the model for the setting. This will return the model instance or raise an exception when not found.

		:return: Model instance
		:raise: NotFound
		"""
		return await SettingModel.get(key=self.key, app=self.app_label)

	def __str__(self):
		return self.name



#!/usr/bin/env python3
"""
Bootstrap script of the PyPlanet Controller.
"""
import os

if __name__ == '__main__':
	# Set the settings options.
	os.environ.setdefault('PYPLANET_SETTINGS_METHOD', 'python')
	os.environ.setdefault('PYPLANET_SETTINGS_MODULE', 'settings')

	# Use YAML for configuration (comment other options and uncomment the following two lines).
	# os.environ.setdefault('PYPLANET_SETTINGS_METHOD', 'yaml')
	# os.environ.setdefault('PYPLANET_SETTINGS_DIRECTORY', 'settings')
	# Will search for two files in the directory: base.yaml and apps.yaml. See http://pypla.net for more information.

	# Use JSON for configuration (comment other options and uncomment the following two lines).
	# os.environ.setdefault('PYPLANET_SETTINGS_METHOD', 'json')
	# os.environ.setdefault('PYPLANET_SETTINGS_DIRECTORY', 'settings')
	# Will search for two files in the directory: base.json and apps.json. See http://pypla.net for more information.

	from pyplanet.core.management import execute_from_command_line
	execute_from_command_line()



import json
import os

from pyplanet.conf.backends.file import FileConfigBackend
from pyplanet.core.exceptions import ImproperlyConfigured


class JsonConfigBackend(FileConfigBackend):
	name = 'json'
	files = ['base.json', 'apps.json']
	required_files = ['base.json', 'apps.json']

	def load(self):
		# Prepare + load directory.
		super().load()

		# Load the files and parse JSON.
		parsed_settings = dict()

		for file_name in self.files:
			try:
				file_path = os.path.join(self.directory, file_name)
				with open(file_path, 'r') as file_handle:
					parsed_settings.update(json.load(file_handle))
			except json.JSONDecodeError as e:
				raise ImproperlyConfigured(
					'Your settings file(s) contain invalid JSON syntax! Please fix and restart! File {}, Error {}'.format(
						file_name,
						str(e)
					)
				)

		# Loop and set in local settings (+ uppercase keys).
		for key, value in parsed_settings.items():
			self.settings[key.upper()] = value



import math

from pyplanet.views.generics.widget import TimesWidgetView
from pyplanet.utils import times


class LiveRankingsWidget(TimesWidgetView):
	widget_x = -160
	widget_y = 70.5
	z_index = 130
	size_x = 38
	size_y = 55.5
	top_entries = 5
	title = 'Live Rankings'

	template_name = 'live_rankings/widget.xml'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_liverankings'

		self.record_amount = 15
		self.format_times = True
		self.display_cpdifference = False

	def get_widget_records(self, player=None):
		list_records = list()

		player_index = len(self.app.current_rankings) + 1
		if player:
			player_record = [x for x in self.app.current_rankings if x['login'] == player.login]
		else:
			player_record = list()

		if len(player_record) > 0:
			# Set player index if there is a record
			player_index = (self.app.current_rankings.index(player_record[0]) + 1)

		records = list(self.app.current_rankings[:self.top_entries])
		if self.app.instance.performance_mode:
			# Performance mode is turned on, get the top of the whole widget.
			records += self.app.current_rankings[self.top_entries:self.record_amount]
			custom_start_index = (self.top_entries + 1)
		else:
			if player_index > len(self.app.current_rankings):
				# No personal record, get the last records
				records_start = (len(self.app.current_rankings) - self.record_amount + self.top_entries)
				# If start of current slice is in the top entries, add more records below
				if records_start < self.top_entries:
					records_start = self.top_entries

				records += list(self.app.current_rankings[records_start:])
				custom_start_index = (records_start + 1)
			else:
				if player_index <= self.top_entries:
					# Player record is in top X, get following records (top entries + 1 onwards)
					records += self.app.current_rankings[self.top_entries:self.record_amount]
					custom_start_index = (self.top_entries + 1)
				else:
					# Player record is not in top X, get records around player record
					# Same amount above the record as below, except when not possible (favors above)
					records_to_fill = (self.record_amount - self.top_entries)
					start_point = ((player_index - math.ceil((records_to_fill - 1) / 2)) - 1)
					end_point = ((player_index + math.floor((records_to_fill - 1) / 2)) - 1)

					# If end of current slice is outside the list, add more records above
					if end_point > len(self.app.current_rankings):
						end_difference = (end_point - len(self.app.current_rankings))
						start_point = (start_point - end_difference)
					# If start of current slice is in the top entries, add more records below
					if start_point < self.top_entries:
						start_point = self.top_entries

					records += self.app.current_rankings[start_point:(start_point + records_to_fill)]
					custom_start_index = (start_point + 1)

		index = 1
		best = None
		for record in records:
			if self.display_cpdifference and index == 1:
				best = record

			list_record = dict()
			list_record['index'] = index
			list_record['bgcolor'] = '00000070'

			if player:
				if player.flow.team_id == 0:
					list_record['bgcolor'] = '7881F2FF'
				if player.flow.team_id == 1:
					list_record['bgcolor'] = 'F05F5FFF'

			list_record['color'] = '$fff'
			if index <= self.top_entries:
				list_record['color'] = '$ff0'
			if index == player_index:
				list_record['color'] = '$0f3'

			list_record['nickname'] = record['nickname']

			if self.format_times:
				list_record['score'] = times.format_time(int(record['score']))
			else:
				list_record['score'] = int(record['score'])

			list_record['cp_difference'] = None

			if self.display_cpdifference and 'cps' in best and 'cps' in record:
				list_record['cp_difference'] = (best['cps'] - record['cps'])

				if index > 1 and not record['finish']:
					# Calculate difference to first player
					best_cp = best['cp_times'][(record['cps'] - 1)]
					current_diff = (record['score'] - best['cp_times'][(record['cps'] - 1)])
					list_record['score'] = '+ ' + times.format_time(int(current_diff))

				if record['finish']:
					list_record['score'] = '$i' + str(list_record['score'])
				elif record['giveup']:
					list_record['score'] = '$iDNF'

			# Points added on finishing in Rounds, Team or Cup modes.
			if 'points_added' in record:
				list_record['points_added'] = record['points_added']
			else:
				list_record['points_added'] = 0

			if index == self.top_entries:
				index = custom_start_index
			else:
				index += 1

			list_records.append(list_record)
		return list_records

	async def get_context_data(self):
		# Determine the Y of the widget. It should be placed under the dedimania widget if it's enabled.
		# Otherwise, take the space below the server information on the top left.
		self.widget_y = 12.5 if self.app.dedimania_enabled else 70.5
		self.record_amount = await self.app.setting_rankings_amount.get_value()
		if self.record_amount < 15:
			self.record_amount = 15

		current_script = await self.app.instance.mode_manager.get_current_script()
		if 'TimeAttack' in current_script or 'TrackMania/TM_TimeAttack_Online' in current_script:
			self.format_times = True
			self.display_cpdifference = False
		elif 'Laps' in current_script or 'TrackMania/TM_Laps_Online' in current_script:
			self.format_times = True
			self.display_cpdifference = True
		else:
			self.format_times = False
			self.display_cpdifference = False

		data = await super().get_context_data()
		if self.app.instance.performance_mode:
			data['times'] = self.get_widget_records()
		return data

	async def get_player_data(self):
		data = await super().get_player_data()

		# If in performance mode, ignore this method.
		if self.app.instance.performance_mode:
			return dict()
		else:
			for player in self.app.instance.player_manager.online:
				data[player.login] = dict(times=self.get_widget_records(player))
			return data



"""
This file contains a glue between core callbacks and desired callbacks.
"""
from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.core.events import Signal, SignalManager


class Callback(Signal):
	"""
	A callback signal is an double signal. Once for the GBX Callback itself (the Gbx callback named). And the destination
	Between those two signals is a sort of `processor` that confirms it into the PyPlanet style objects.

	For example, a player connect will result in a player database object instead of the plain Maniaplanet payload.
	This will make it possible to develop your app as fast as possible, without any overhead and make it better
	with callback payload changes!
	"""
	def __init__(self, call, namespace, code, target=None):
		"""
		Shortcut for registering two signals, one is the raw signal and the second one is the parsed and structured
		output signal. This also glues the two together.

		:param call:
		:param namespace:
		:param code:
		:param target:
		"""
		# Initiate destination signal (ourself).
		super().__init__(code=code, namespace=namespace, process_target=target)

		# Initiate raw signal, the raw gbx/script callback.
		self.raw_signal = Signal(code=call, namespace='raw')
		self.raw_signal.register(self.glue, weak=False)

		SignalManager.register_signal(self.raw_signal, app=None, callback=True)
		SignalManager.register_signal(self, app=None)

	async def glue(self, signal, source, **kwargs):
		"""
		The glue method converts the source signal (gbx callback) into the pyplanet signal.
		"""
		return await self.send_robust(source)


async def handle_generic(source, signal, **kwargs):
	"""
	The handle_generic is a simple handle (`processing glue`) for just forwarding the payload from the maniaplanet
	server into the signal payload.
	"""
	if not isinstance(source, dict):
		source = dict(raw=source)
	if 'login' in source:
		try:
			if source['login']:
				source['player'] = await Player.get_by_login(source['login'])
			else:
				source['player'] = None
		except:
			source['player'] = None

	return source



"""
Donation Toolbar View.
"""

from pyplanet.views.generics.widget import WidgetView


class DonationToolbarWidget(WidgetView):
	template_name = 'transactions/donation_toolbar.xml'

	widget_x = -109.5
	widget_y = 61
	z_index = 130
	game = None

	def __init__(self, app):
		"""
		Initiate the toolbar view.

		:param app: App instance.
		:type app: pyplanet.apps.contrib.transactions.Transactions
		"""
		super().__init__(app.context.ui)

		self.app = app
		self.id = 'donate_widget'

		self.subscribe('bar_button_100P', self.action_donate(100))
		self.subscribe('bar_button_500P', self.action_donate(500))
		self.subscribe('bar_button_1000P', self.action_donate(1000))
		self.subscribe('bar_button_2000P', self.action_donate(2000))
		self.subscribe('bar_button_5000P', self.action_donate(5000))

	async def display(self, player=None, **kwargs):
		if not self.game:
			self.game = self.app.instance.game.game
			if self.game == 'sm':
				self.widget_x = -145

		return await super().display(player, **kwargs)

	async def get_context_data(self):
		data = await super().get_context_data()
		data['game'] = self.app.instance.game.game
		return data

	def action_donate(self, planets):
		async def donate(player, *args, **kwargs):
			return await self.app.instance.command_manager.execute(player, '/donate {}'.format(planets))
		return donate



import asynctest
import datetime

from pyplanet.contrib.setting import Setting
from pyplanet.contrib.setting.exceptions import SerializationException
from pyplanet.core import Controller


class TestSettings(asynctest.TestCase):
	async def test_registration(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()

		test_1 = Setting('test1', 'Test 1', Setting.CAT_GENERAL, type=str)
		await instance.setting_manager.register(
			test_1
		)
		setting_list = await instance.setting_manager.get_all()
		assert test_1 in setting_list

	async def test_saving(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()

		test_1 = Setting('test1', 'Test 1', Setting.CAT_GENERAL, type=str)
		await instance.setting_manager.register(
			test_1
		)

		expected = 'test1'
		await test_1.set_value('test1')
		cached = await test_1.get_value(refresh=False)
		real = await test_1.get_value(refresh=True)

		assert cached == expected
		assert real == expected

	async def test_validating(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()

		test_1 = Setting('test1', 'Test 1', Setting.CAT_GENERAL, type=str)
		await instance.setting_manager.register(
			test_1
		)

		with self.assertRaises(SerializationException) as context:
			await test_1.set_value(True)
		assert isinstance(context.exception, SerializationException)

		with self.assertRaises(SerializationException) as context:
			await test_1.set_value(1)
		assert isinstance(context.exception, SerializationException)

		with self.assertRaises(SerializationException) as context:
			await test_1.set_value(float(55))
		assert isinstance(context.exception, SerializationException)

		with self.assertRaises(SerializationException) as context:
			await test_1.set_value(dict())
		assert isinstance(context.exception, SerializationException)

		with self.assertRaises(SerializationException) as context:
			await test_1.set_value(list())
		assert isinstance(context.exception, SerializationException)

		value = test_1.get_value(refresh=True)
		assert value is not True



import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.apps.core.trackmania import callbacks as tm_signals

from .view import BestCpTimesWidget
from .view import CpTimesListView


class BestCpTimes(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next']
	app_dependencies = ['core.maniaplanet', 'core.trackmania']
	# mode_dependencies = ['TimeAttack']

	best_cp_times = []

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.number_of_checkpoints = None
		self.best_cp_times = []  # List of PlayerCP Objects
		self.widget = None

	async def on_start(self):
		self.context.signals.listen(tm_signals.waypoint, self.player_cp)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.map.map_begin, self.map_begin)
		self.context.signals.listen(mp_signals.map.map_start__end, self.map_end)
		self.best_cp_times.clear()
		self.widget = BestCpTimesWidget(self)
		asyncio.ensure_future(self.widget.display())

	# When a player passes a CP
	async def player_cp(self, player, raw, *args, **kwargs):
		cpnm = int(raw['checkpointinlap'])
		laptime = int(raw['laptime'])

		# Ignore invalid times (cp time will be 0)
		if laptime == 0:
			return

		pcp = PlayerCP(player, cpnm + 1, laptime)
		if not self.best_cp_times:
			self.best_cp_times.append(pcp)
		else:
			added = False
			for idx, current_cp in enumerate(self.best_cp_times):
				if pcp.cp < current_cp.cp:
					self.best_cp_times.insert(idx, pcp)
					added = True
					break
				elif pcp.cp == current_cp.cp:
					if laptime < current_cp.time:
						self.best_cp_times[idx] = pcp
					added = True
					break
			if not added:
				self.best_cp_times.append(pcp)
		await self.widget.display()

	# When the map starts
	async def map_begin(self, *args, **kwargs):
		self.best_cp_times.clear()
		await self.widget.display()

	# When the map ends. This is needed to not keep the old CP data when the map is restarted.
	async def map_end(self, *args, **kwargs):
		self.best_cp_times.clear()
		await self.widget.display()

	# When a player connects
	async def player_connect(self, player, **kwargs):
		await self.widget.display(player)


# PlayerCP Event mapping a player to a cp and a time
class PlayerCP:
	def __init__(self, player, cp=0, time=0):
		self.player = player
		self.cp = cp
		self.time = time



import random

from tests.integration import ControllerTestCase


class TestMapManager(ControllerTestCase):

	async def test_map_list(self):
		real_list = await self.instance.gbx.execute('GetMapList', -1, 0)
		assert len(real_list) == len(self.instance.map_manager.maps)

	async def test_map_juke(self):
		if len(self.instance.map_manager.maps) <= 1:
			raise Exception('Test server should contain more than 1 map!')
		while True:
			next_map = random.choice(list(self.instance.map_manager.maps))
			if next_map.uid != self.instance.map_manager.current_map.uid:
				break

		# Set next map
		await self.instance.map_manager.set_next_map(next_map)
		assert self.instance.map_manager.next_map == next_map
		map_info = await self.instance.gbx.execute('GetNextMapInfo')
		assert map_info['UId'] == next_map.uid
	#
	# async def test_map_jump(self):
	# 	if len(self.instance.map_manager.maps) <= 1:
	# 		raise Exception('Test server should contain more than 1 map!')
	# 	while True:
	# 		jump_map = random.choice(self.instance.map_manager.maps)
	# 		if jump_map.uid != self.instance.map_manager.current_map.uid:
	# 			break
	#
	# 	# Skip to next map.
	# 	await self.instance.map_manager.set_current_map(jump_map)
	# 	assert self.instance.map_manager.next_map == jump_map



import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.contrib.local_records.views import LocalRecordsListView, LocalRecordsWidget
from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.contrib.command import Command
from pyplanet.contrib.setting import Setting

from pyplanet.apps.core.trackmania import callbacks as tm_signals
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

from pyplanet.utils import times
from pyplanet.utils.log import handle_exception

from .models import LocalRecord


class LocalRecords(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next']
	app_dependencies = ['core.maniaplanet', 'core.trackmania']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.lock = asyncio.Lock()

		self.current_records = []
		self.widget = None

		self.setting_chat_announce = Setting(
			'chat_announce', 'Minimum index for chat announce', Setting.CAT_BEHAVIOUR, type=int,
			description='Minimum record index needed for public new record/recordchange announcement (0 for disable).',
			default=50
		)

		self.setting_record_limit = Setting(
			'record_limit', 'Local Records limit', Setting.CAT_BEHAVIOUR, type=int,
			description='Limit for the amount of Local Records displayed in-game (0 for disable).',
			default=100
		)

	async def on_start(self):
		# Register commands
		await self.instance.command_manager.register(
			Command(command='records', target=self.show_records_list,
					description='Displays the complete list of local records on the current map.'),
			Command(
				'localcps', target=self.command_localcps, description='Compares your local record checkpoints with another record.'
			).add_param('record', required=False, type=int, help='Custom record rank to compare with. Defaults to 1.', default=1)
		)

		# Register signals
		self.context.signals.listen(mp_signals.map.map_begin, self.map_begin)
		self.context.signals.listen(tm_signals.finish, self.player_finish)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)

		# Register settings
		await self.context.setting.register(self.setting_chat_announce, self.setting_record_limit)

		# Register permissions
		await self.instance.permission_manager.register('manage_records', 'Manage records', app=self, min_level=3)

		# Load initial data.
		await self.refresh_locals()
		await self.chat_current_record()

		if self.widget is None:
			self.widget = LocalRecordsWidget(self)

		await self.widget.display()

	async def load_map_locals(self, map=None):
		if map:
			map.local = await self.get_map_record(map)
		else:
			maps = {m.id: m for m in self.instance.map_manager.maps}
			map_locals = dict()

			# Fetch all.
			rows = await LocalRecord.execute(
				LocalRecord.select().where(
					LocalRecord.map_id << list(maps.keys())
				)
			)

			# Group by map.
			for row in rows:
				if row.map_id not in map_locals:
					map_locals[row.map_id] = list()
				map_locals[row.map_id].append(row)

			# Map local stats.
			for map_id, local_list in map_locals.items():
				maps[map_id].local = dict(
					record_count=len(local_list),
					first_record=local_list[0] if len(local_list) > 0 else None
				)

			del map_locals
			del maps

	async def get_map_record(self, map=None):
		if not map:
			map = self.instance.map_manager.current_map

		record_list = await LocalRecord.objects.execute(
			LocalRecord.select(LocalRecord, Player)
				.join(Player)
				.where(LocalRecord.map_id == map.get_id())
				.order_by(LocalRecord.score.asc())
		)

		return {
			'record_count': len(record_list),
			'first_record': record_list[0] if len(record_list) > 0 else None
		}

	async def get_player_record_for_map(self, map, player):
		record_list = await LocalRecord.objects.execute(
			LocalRecord.select(LocalRecord)
				.where(LocalRecord.player_id == player.get_id())
				.where(LocalRecord.map_id == map.get_id())
		)

		return record_list

	async def get_player_record_and_rank_for_map(self, map, player):
		record_list = await LocalRecord.objects.execute(
			LocalRecord.select(LocalRecord, Player)
				.join(Player)
				.where(LocalRecord.map_id == map.get_id())
				.order_by(LocalRecord.score.asc())
		)

		rank = 1
		for rec in record_list:
			if rec.player == player:
				return rank, rec
			rank += 1
		return None, None

	async def get_local(self, id):
		return await LocalRecord.get(id=id)

	async def refresh(self):
		await self.refresh_locals()
		if self.widget:
			await self.widget.refresh()

	async def delete_record(self, record):
		await LocalRecord.execute(
			LocalRecord.delete().where(LocalRecord.id == record.get_id())
		)

	async def refresh_locals(self):
		record_list = await LocalRecord.objects.execute(
			LocalRecord.select(LocalRecord, Player)
				.join(Player)
				.where(LocalRecord.map_id == self.instance.map_manager.current_map.get_id())
				.order_by(LocalRecord.score.asc())
		)
		self.current_records = list(record_list)

	async def show_records_list(self, player, data = None, **kwargs):
		"""
		Show record list view to player.

		:param player: Player instance.
		:param data: -
		:param kwargs: -
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: view instance or nothing when there are no records.
		"""
		if not len(self.current_records):
			message = '$i$f00There are currently no records on this map!'
			await self.instance.chat(message, player)
			return

		view = LocalRecordsListView(self)
		await view.display(player=player.login)
		return view

	async def map_begin(self, map):
		await self.refresh_locals()

		if self.widget is None:
			self.widget = LocalRecordsWidget(self)

		await asyncio.gather(
			self.chat_current_record(),
			self.widget.display()
		)

	async def player_connect(self, player, is_spectator, source, signal):
		if self.widget:
			await self.widget.display(player=player)

	async def player_finish(self, player, race_time, lap_time, cps, flow, raw, **kwargs):
		record_limit = await self.setting_record_limit.get_value()
		chat_announce = await self.setting_chat_announce.get_value()
		async with self.lock:
			current_records = [x for x in self.current_records if x.player.login == player.login]
			score = lap_time

			previous_index = None
			previous_time = None

			if len(current_records) > 0:
				current_record = current_records[0]
				if score > current_record.score:
					# No improvement, ignore
					return

				# Temporary make index + time local for the messages.
				previous_index = self.current_records.index(current_record) + 1
				previous_time = current_record.score

				# If equal, only show message.
				if score == current_record.score and (record_limit == 0 or previous_index <= record_limit):
					message = '$fff{}$z$s$0f3 equalled the $fff{}.$0f3 Local Record: $fff\uf017 {}$0f3.'.format(
						player.nickname, previous_index, times.format_time(score)
					)

					if chat_announce >= previous_index:
						return await self.instance.chat(message)
					elif chat_announce != 0:
						return await self.instance.chat(message, player.login)

			else:
				current_record = LocalRecord(
					map=self.instance.map_manager.current_map,
					player=player,
				)

			# Set details (score + cps times).
			current_record.score = score
			current_record.checkpoints = ','.join([str(cp) for cp in cps])

			# Add to list when it's a new record!
			if current_record.get_id() is None:
				self.current_records.append(current_record)

			# (Re)sort the record list.
			self.current_records.sort(key=lambda x: x.score)
			new_index = self.current_records.index(current_record) + 1

			if new_index == 1:
				map = next((m for m in self.instance.map_manager.maps if m.uid == self.instance.map_manager.current_map.uid), None)
				if map is not None:
					map.local = {'record_count': len(self.current_records), 'first_record': current_record}

		# Prepare messages.
		if previous_index is not None and (record_limit == 0 or previous_index <= record_limit):
			if new_index < previous_index:
				message = '$fff{}$z$s$0f3 gained the $fff{}.$0f3 Local Record: $fff\uf017 {}$0f3 ($fff{}.$0f3 $fff-{}$0f3).'.format(
					player.nickname, new_index, times.format_time(score), previous_index,
					times.format_time((previous_time - score))
				)
			else:
				message = '$fff{}$z$s$0f3 improved the $fff{}.$0f3 Local Record: $fff\uf017 {}$0f3 ($fff-{}$0f3).'.format(
					player.nickname, new_index, times.format_time(score),
					times.format_time((previous_time - score))
				)
		else:
			message = '$fff{}$z$s$0f3 drove the $fff{}.$0f3 Local Record: $fff\uf017 {}$0f3.'.format(
				player.nickname, new_index, times.format_time(score)
			)

		# Save to database (but don't wait for it).
		try:
			asyncio.ensure_future(current_record.save())
		except Exception as e:
			# To investigate #283.
			handle_exception(e, __name__, 'player_finish', extra_data={
				'own_records': current_records,
				'own_record': current_record
			})

		if self.widget is None:
			self.widget = LocalRecordsWidget(self)

		coros = [self.widget.display()]
		if record_limit == 0 or new_index <= record_limit:
			if chat_announce >= new_index:
				coros.append(self.instance.chat(message))
			elif chat_announce != 0:
				coros.append(self.instance.chat(message, player))
		await asyncio.gather(*coros)

		# Reload map referenced information
		asyncio.ensure_future(self.load_map_locals(map=self.instance.map_manager.current_map))

	async def chat_current_record(self):
		record_limit = await self.setting_record_limit.get_value()
		if record_limit > 0:
			records_amount = len(self.current_records[:record_limit])
		else:
			records_amount = len(self.current_records)

		if records_amount > 0:
			first_record = self.current_records[0]
			message = '$0f3Current Local Record: $fff\uf017 {}$z$s$0f3 by $fff{}$z$s$0f3 ($fff{}$0f3 records)'.format(
				times.format_time(first_record.score), first_record.player.nickname, records_amount
			)
			calls = list()
			calls.append(self.instance.chat(message))
			for player in self.instance.player_manager.online:
				calls.append(self.chat_personal_record(player, record_limit))
			await self.instance.gbx.multicall(*calls)
		else:
			message = '$0f3There is no Local Record on this map yet.'
			await self.instance.chat(message)

	def chat_personal_record(self, player, record_limit):
		if record_limit > 0:
			records = self.current_records[:record_limit]
		else:
			records = self.current_records

		record = [x for x in records if x.player_id == player.get_id()]

		if len(record) > 0:
			message = '$0f3You currently hold the $fff{}.$0f3 Local Record: $fff\uf017 {}'.format(
				self.current_records.index(record[0]) + 1, times.format_time(record[0].score)
			)
			return self.instance.chat(message, player)
		else:
			message = '$0f3You don\'t have a Local Record on this map yet.'
			return self.instance.chat(message, player)

	async def command_localcps(self, player, data, *args, **kwargs):
		"""
		Compare local checkpoints from your record to another record, defaults to 1st record.

		:param player: Player instance
		:param data: Arguments data
		:param args: *
		:param kwargs: **
		:return:
		"""
		async with self.lock:
			record = [x for x in self.current_records if x.player_id == player.get_id()]

			if data.record > len(self.current_records):
				message = '$0b3There is no record for rank {}!'.format(data.record)
				return await self.instance.chat(message, player)

			compare_record = self.current_records[data.record - 1]

			record_index = (self.current_records.index(record[0]) + 1) if len(record) else None
			compare_index = (self.current_records.index(compare_record) + 1)

		view = views.LocalRecordCpCompareListView(
			self, record[0] if len(record) else None, record_index, compare_record, compare_index
		)
		await view.display(player)



"""
PyPlanet Instance Module

This module holds the main instance class of the PyPlanet system.
"""
import os
import asyncio
import logging
import traceback

from pyplanet.utils.analytics import Analytics
from .controller import Controller as _Controller

from pyplanet import __version__ as version

from pyplanet.apps import Apps
from pyplanet.conf import settings
from pyplanet.core import signals
from pyplanet.core.events import SignalManager
from pyplanet.core.db.database import Database
from pyplanet.core.game import Game
from pyplanet.core.gbx import GbxClient
from pyplanet.core.exceptions import ImproperlyConfigured
from pyplanet.core.storage.storage import Storage
from pyplanet.core.ui import GlobalUIManager
from pyplanet.utils import memleak, releases

from pyplanet.contrib.map import MapManager
from pyplanet.contrib.player import PlayerManager
from pyplanet.contrib.command import CommandManager
from pyplanet.contrib.permission import PermissionManager
from pyplanet.contrib.setting import GlobalSettingManager
from pyplanet.contrib.mode import ModeManager
from pyplanet.contrib.chat import ChatManager

logger = logging.getLogger(__name__)


class Instance:
	"""
	Controller Instance. The very base of the controller, containing class instances of all core components.

	:ivar process_name: Process and pool name.
	:ivar loop: AsyncIO Event Loop.
	:ivar game: Game Information class.
	:ivar apps: Apps component.
	:ivar gbx: Gbx component.
	:ivar db: Database component.
	:ivar storage: Storage component.
	:ivar signals: Signal Manager (global). Please use the APP context Signal Manager instead!
	:ivar ui_manager: UI Manager (global). Please use the APP context UI Manager instead!

	:ivar map_manager: Contrib: Map Manager.
	:ivar player_manager: Contrib: Player Manager.
	:ivar permission_manager: Contrib: Permission Manager.
	:ivar command_manager: Contrib: Command Manager.
	:ivar setting_manager: Contrib: Setting Manager. Please use the APP context setting manager instead!
	:ivar mode_manager: Contrib. Mode Manager.
	"""

	def __init__(self, process_name):
		"""
		The actual instance of the controller.

		:param process_name: EnvironmentProcess class specific for this process.
		:type process_name: str
		"""
		# Initiate all the core components.
		self.process_name = 		process_name
		self.loop = 				asyncio.get_event_loop()
		self.game =					Game

		self.gbx = 					GbxClient.create_from_settings(self, settings.DEDICATED[self.process_name])
		self.db = 					Database.create_from_settings(self, settings.DATABASES[self.process_name])
		self.storage =				Storage.create_from_settings(self, settings.STORAGE[self.process_name])
		self.signals =				SignalManager
		self.ui_manager =			GlobalUIManager(self)
		self.apps = 				Apps(self)

		# Contrib components.
		self.map_manager =				MapManager(self)
		self.player_manager =			PlayerManager(self)
		self.permission_manager =		PermissionManager(self)
		self.command_manager =			CommandManager(self)
		self.setting_manager =			GlobalSettingManager(self)
		self.mode_manager =				ModeManager(self)
		self.chat_manager = self.chat = ChatManager(self)

		# Populate apps.
		self.apps.populate(settings.MANDATORY_APPS, in_order=True)
		try:
			self.apps.populate(settings.APPS[self.process_name])
		except KeyError as e:
			raise ImproperlyConfigured(
				'One of the pool names doesn\'t reflect into the APPS setting! You must '
				'declare the apps per pool! ({})'.format(str(e))
			)

	def start(self, run_forever=True):  # pragma: no cover
		"""
		Start wrapper.
		"""
		try:
			# Start memleak checker.
			memleak.checker.start()

			# Initiate instance.
			self.loop.run_until_complete(self._start())

			# Run forever.
			if run_forever:
				self.loop.run_forever()
		except KeyboardInterrupt:  # pragma: no branch
			pass
		except Exception as e:
			logger.exception(e)
			raise
		finally:
			self.stop()

	def stop(self):  # pragma: no cover
		"""
		Stop all the instance apps and managers.
		"""
		try:
			self.loop.run_until_complete(self._stop())
		except Exception as e:
			logger.exception(e)
			raise
		os._exit(0)

	@property
	def performance_mode(self):
		"""
		Gives back a boolean, True if we are in performance mode.

		:return: Performance mode boolean.
		"""
		return self.player_manager.performance_mode

	async def __fire_signal(self, signal):  # pragma: no cover
		"""
		Fire signal with given name to all listeners.

		:param signal: Signal to fire on.
		:type signal: pyplanet.core.events.dispatcher.Signal
		"""
		await signal.send(dict(instance=self))

	async def _start(self):  # pragma: no cover
		"""
		The start coroutine is executed when the process is ready to create connection to the gbx protocol, database,
		other services and finally start the apps.
		"""
		# Make sure we start the Gbx connection, authenticate, set api version and stuff.
		await self.__fire_signal(signals.pyplanet_start_gbx_before)
		await self.gbx.connect()
		await self.__fire_signal(signals.pyplanet_start_gbx_after)

		# Initiate the database connection, discover apps assets,models etc.
		await self.__fire_signal(signals.pyplanet_start_db_before)
		await self.db.connect()				# Connect and initial state.
		await self.apps.discover() 			# Discover apps models.
		await self.db.initiate() 			# Execute migrations and initial tasks.
		await self.apps.check(True)    		# Check for incompatible apps and remove them.
		await self.apps.init()				# Initiate apps
		await self.ui_manager.on_start()    # Initiate UI manager.
		await self.__fire_signal(signals.pyplanet_start_db_after)

		# Start the core contribs.
		await self.setting_manager.on_start()
		await self.map_manager.on_start()
		await self.player_manager.on_start()
		await self.permission_manager.on_start()
		await self.command_manager.on_start()
		await self.mode_manager.on_start()
		await self.chat_manager.on_start()

		# Start the apps, call the on_ready, resulting in apps user logic to be started.
		await self.print_header()
		await self.__fire_signal(signals.pyplanet_start_apps_before)
		await self.apps.start()
		await self.__fire_signal(signals.pyplanet_start_apps_after)
		await self.print_footer()

		# Utils.
		await Analytics.start(self)

		# Finish signalling and send finish signal.
		await self.signals.finish_start()
		await self.__fire_signal(signals.pyplanet_start_after)

	async def _stop(self):
		"""
		The stop coroutine is executed when the process exits with the SIGINT signal.
		"""
		await self.apps.stop()

	async def print_header(self):  # pragma: no cover
		await self.chat.execute(
			self.chat('', raw=True),
			self.chat('$fff$o$w⏳$z$fff Loading...', raw=True)
		)

	async def print_footer(self):  # pragma: no cover
		await self.chat(
			'\uf1e6 $o$FD4Py$369Planet$z$o$s$fff v{}, {}\uf013 $z$s $369|$FD4 '
			'$l[http://pypla.net]Site$l $369|$FD4 '
			'$l[https://github.com/PyPlanet]Github$l $369|$FD4 '
			'$l[http://pypla.net]Docs$l'.format(version, len(self.apps.apps)),
			raw=True
		)

		if self.game.game == 'tmnext':
			await self.chat('$f00$oPyPlanet: $o$nNewly support for this Trackmania version. Expect issues'
							', please report to PyPlanet.')

		try:
			asyncio.ensure_future(releases.UpdateChecker.init_checker(self))
		except:
			pass  # Completely ignore errors while checking for the latest version.


Controller = _Controller
"""
Controller access point to prevent circular imports. This is a lazy provided way to get the instance from anywhere!
:type Controller: pyplanet.core.Controller
:type: pyplanet.core.Controller
"""



import asynctest

from pyplanet.contrib.chat.exceptions import ChatException
from pyplanet.contrib.chat.query import ChatQuery
from pyplanet.core import Controller


class TestChat(asynctest.TestCase):
	async def test_simple(self):
		instance = Controller.prepare(name='default').instance
		chat = instance.chat.prepare()
		assert isinstance(chat, ChatQuery)

		chat.to_all().message('Test')
		assert len(chat.get_formatted_message()) > len('Test')  # Must include the prefix!!

	async def test_to_logins(self):
		instance = Controller.prepare(name='default').instance
		await instance.db.connect()
		await instance.apps.discover()
		await instance.db.initiate()

		from pyplanet.apps.core.maniaplanet.models import Player

		try:
			player1 = await Player.get(login='sample-1')
		except:
			player1 = await Player.create(login='sample-1', nickname='sample-1', level=Player.LEVEL_MASTER)

		try:
			player2 = await Player.get(login='sample-2')
		except:
			player2 = await Player.create(login='sample-2', nickname='sample-2', level=Player.LEVEL_PLAYER)

		# Try list.
		chat = instance.chat.prepare().to_players([player1, player2])
		assert 'sample-1' in chat._logins and 'sample-2' in chat._logins

		# Try unpacked list.
		chat = instance.chat.prepare().to_players(*[player1, player2])
		assert 'sample-1' in chat._logins and 'sample-2' in chat._logins

		# Try login list.
		chat = instance.chat.prepare().to_players([player1.login, player2.login])
		assert 'sample-1' in chat._logins and 'sample-2' in chat._logins

		# Try login unpacked list.
		chat = instance.chat.prepare().to_players(*[player1.login, player2.login])
		assert 'sample-1' in chat._logins and 'sample-2' in chat._logins

	async def test_query_conversion(self):
		instance = Controller.prepare(name='default').instance
		# MOCK:
		instance.gbx.gbx_methods = ['ChatSendServerMessageToLogin', 'ChatSendServerMessage']

		chat = instance.chat.prepare().to_players(['test-1', 'test-2']).message('Test')
		assert chat.gbx_query.method == 'ChatSendServerMessageToLogin'
		assert len(chat.gbx_query.args) == 2

		chat = instance.chat.prepare().message('Test')
		assert chat.gbx_query.method == 'ChatSendServerMessage'
		assert len(chat.gbx_query.args) == 1

		chat = instance.chat.prepare().to_all().message('Test')
		assert chat.gbx_query.method == 'ChatSendServerMessage'
		assert len(chat.gbx_query.args) == 1

	async def test_short_syntax(self):
		instance = Controller.prepare(name='default').instance
		# MOCK:
		instance.gbx.gbx_methods = ['ChatSendServerMessageToLogin', 'ChatSendServerMessage']

		prepared = instance.chat('Test')
		assert isinstance(prepared, ChatQuery)



"""
Trackmania app component.
"""
from pyplanet.apps.core.statistics.models import Score
from pyplanet.apps.core.statistics.views.dashboard import StatsDashboardView
from pyplanet.apps.core.statistics.views.records import TopSumsView
from pyplanet.apps.core.statistics.views.score import StatsScoresListView, CheckpointComparisonView
from pyplanet.apps.core.trackmania.callbacks import finish
from pyplanet.contrib.command import Command


class TrackmaniaComponent:
	def __init__(self, app):
		"""
		Initiate trackmania statistics component.

		:param app: App config instance
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		self.app = app

	async def on_init(self):
		pass

	async def on_start(self):
		# Listen to signals.
		self.app.context.signals.listen(finish, self.on_finish)

		# Register commands.
		await self.app.instance.command_manager.register(
			# Command('stats', target=self.open_stats),
			Command('topsums', target=self.topsums, description='Displays a list of top record players.'),
			Command(command='scoreprogression', aliases=['progression'], target=self.open_score_progression,
					description='Displays your time/score progression on the current map.'),
			Command(command='cpcomparison', aliases=['cp'], target=self.open_cp_comparison,
					description='Compares your checkpoints times with the local record and the ideal checkpoints.'),
		)

	async def on_finish(self, player, race_time, lap_time, cps, flow, raw, **kwargs):
		# Register the score of the player.
		await Score(
			player=player,
			map=self.app.instance.map_manager.current_map,
			score=race_time,
			checkpoints=','.join([str(cp) for cp in cps])
		).save()

	async def open_stats(self, player, **kwargs):
		view = StatsDashboardView(self.app, self.app.context.ui, player)
		await view.display()

	async def open_score_progression(self, player, **kwargs):
		view = StatsScoresListView(self.app, player)
		await view.display(player)

	async def open_cp_comparison(self, player, **kwargs):
		view = CheckpointComparisonView(self.app, player)
		await view.display(player)

	async def topsums(self, player, *args, **kwargs):
		await self.app.instance.chat('$0f3Loading Top Record Players ...', player)
		view = TopSumsView(self.app, player, await self.app.processor.get_topsums())
		await view.display(player)



import asyncio
import functools
import pkgutil
import os
import sys

from collections import defaultdict
from importlib import import_module

from pyplanet import __version__ as version
from pyplanet.conf import settings
from pyplanet.core.exceptions import ImproperlyConfigured
from pyplanet.core.management.base import CommandParser, CommandError, handle_default_options, BaseCommand
from pyplanet.core.management.color import color_style


def find_commands(management_dir):
	command_dir = os.path.join(management_dir, 'commands')
	return [name for _, name, is_pkg in pkgutil.iter_modules([command_dir])]


@functools.lru_cache(maxsize=None)
def get_commands():
	"""
	Return a dictionary mapping command names to their callback applications.

	Look for a management.commands package in pyplanet.core, and in each
	installed application -- if a commands package exists, register all
	commands in that package.

	Core commands are always included. If a settings module has been
	specified, also include user-defined commands.

	The dictionary is in the format {command_name: app_name}. Key-value
	pairs from this dictionary can then be used in calls to
	load_command_class(app_name, command_name).

	If a specific version of a command must be loaded (e.g., with the
	startapp command), the instantiated module can be placed in the
	dictionary in place of the application name.

	The dictionary is cached on the first call and reused on subsequent
	calls.
	"""
	commands = {name: 'pyplanet.core' for name in find_commands(__path__[0])}

	if not settings.configured:
		return commands

	# TODO: Implement later.
	# for app_config in reversed(list(apps.get_app_configs())):
	# 	path = os.path.join(app_config.path, 'management')
	# 	commands.update({name: app_config.name for name in find_commands(path)})

	return commands


def load_command_class(app_name, name):
	"""
	Given a command name and an application name, return the Command
	class instance. Allow all errors raised by the import process
	(ImportError, AttributeError) to propagate.
	"""
	module = import_module('{}.management.commands.{}'.format(app_name, name))
	return module.Command()


class ManagementUtility:
	def __init__(self, argv):
		self.argv = argv or sys.argv[:]
		self.prog_name = os.path.basename(self.argv[0])
		self.version = version
		self.loop = asyncio.get_event_loop()
		self.settings_exception = None

		self.commands = find_commands(__path__[0])

		settings._setup(optional_loading=True)

	def fetch_command(self, subcommand):
		"""
		Try to fetch the given subcommand, printing a message with the
		appropriate command called from the command line (usually "pyplanet" or "manage.py") if it can't be found.
		"""
		# Get commands outside of try block to prevent swallowing exceptions
		commands = get_commands()
		try:
			app_name = commands[subcommand]
		except KeyError:
			if os.environ.get('PYPLANET_SETTINGS_MODULE'):
				# If `subcommand` is missing due to misconfigured settings, the
				# following line will retrigger an ImproperlyConfigured exception
				# (get_commands() swallows the original one) so the user is
				# informed about it.
				_ = settings.APPS
			else:
				sys.stderr.write('No PyPlanet settings specified.\n')
			sys.stderr.write(
				'Unknown command: {}\nType \'{} help\' for usage.\n'.format(subcommand, self.prog_name)
			)
			sys.exit(1)
		if isinstance(app_name, BaseCommand):
			# If the command is already loaded, use it directly.
			klass = app_name
		else:
			klass = load_command_class(app_name, subcommand)
		return klass

	def execute(self):
		try:
			subcommand = self.argv[1]
		except:
			subcommand = 'help'

		# Check for platform requirements. (Python version)
		if sys.version_info[0] < 3 or sys.version_info[1] <= 5:
			print('WARNING')
			print('WARNING: Your Python version is not officially supported by PyPlanet, please upgrade your Python!')
			print('WARNING')
		if sys.version_info[0] > 3 or sys.version_info[1] > 8:
			print('WARNING')
			print('WARNING: Your Python version is untested and most likely not working with PyPlanet!')
			print('WARNING')

		# Root parser.
		parser = CommandParser(None, usage='%(prog)s subcommand [options] [args]', add_help=False)
		parser.add_argument('--settings')
		parser.add_argument('--pool')
		parser.add_argument('--pythonpath')
		parser.add_argument('args', nargs='*')
		try:
			options, args = parser.parse_known_args(self.argv[2:])
			handle_default_options(options)
		except CommandError:
			pass # Ignore any option errors at this point.

		try:
			settings.APPS
		except ImproperlyConfigured as e:
			self.settings_exception = e

		# TODO: Load some parts of the app (with the provided pool or default pool) so we can load commands from apps.

		self.autocomplete()

		if subcommand == 'help':
			if '--commands' in args:
				sys.stdout.write(self.main_help_text(commands_only=True) + '\n')
			elif len(options.args) < 1:
				sys.stdout.write(self.main_help_text() + '\n')
			else:
				self.fetch_command(options.args[0]).print_help(self.prog_name, options.args[0])
			# Special-cases: We want 'pyplanet --version' and
			# 'pyplanet --help' to work, for backwards compatibility.
		elif subcommand == 'version' or self.argv[1:] == ['--version']:
			sys.stdout.write(self.version + '\n')
		elif self.argv[1:] in (['--help'], ['-h']):
			sys.stdout.write(self.main_help_text() + '\n')
		else:
			self.fetch_command(subcommand).run_from_argv(self.argv)

	def main_help_text(self, commands_only=False):
		"""Return the script's main help text, as a string."""
		if commands_only:
			usage = sorted(get_commands().keys())
		else:
			usage = [
				"",
				"Type '{} help <subcommand>' for help on a specific subcommand.".format(self.prog_name),
				"",
				"Available subcommands:",
				]

			commands_dict = defaultdict(lambda: [])
			for name, app in get_commands().items():
				if app == 'pyplanet.core':
					app = 'pyplanet'
				else:
					app = app.rpartition('.')[-1]
				commands_dict[app].append(name)

			style = color_style()

			for app in sorted(commands_dict.keys()):
				usage.append("")
				usage.append(style.NOTICE("[{}]".format(app)))
				for name in sorted(commands_dict[app]):
					usage.append("    {}".format(name))

			# Output an extra note if settings are not properly configured
			if self.settings_exception is not None:
				usage.append(style.NOTICE(
					'\n'
					'Note that only PyPlanet core commands are listed '
					'as settings are not properly configured (error: {}).'.format(self.settings_exception)
				))

		return '\n'.join(usage)

	def autocomplete(self):
		"""
		Output completion suggestions for BASH.

		The output of this function is passed to BASH's `COMREPLY` variable and
		treated as completion suggestions. `COMREPLY` expects a space
		separated string as the result.

		The `COMP_WORDS` and `COMP_CWORD` BASH environment variables are used
		to get information about the cli input. Please refer to the BASH
		man-page for more information about this variables.

		Subcommand options are saved as pairs. A pair consists of
		the long option string (e.g. '--exclude') and a boolean
		value indicating if the option requires arguments. When printing to
		stdout, an equal sign is appended to options which require arguments.

		Note: If debugging this function, it is recommended to write the debug
		output in a separate file. Otherwise the debug output will be treated
		and formatted as potential completion suggestions.
		"""
		# Don't complete if user hasn't sourced bash_completion file.
		if 'PYPLANET_AUTO_COMPLETE' not in os.environ:
			return

		cwords = os.environ['COMP_WORDS'].split()[1:]
		cword = int(os.environ['COMP_CWORD'])

		try:
			curr = cwords[cword - 1]
		except IndexError:
			curr = ''

		subcommands = list(get_commands()) + ['help']
		options = [('--help', False)]

		# subcommand
		if cword == 1:
			print(' '.join(sorted(filter(lambda x: x.startswith(curr), subcommands))))
		# subcommand options
		# special case: the 'help' subcommand has no options
		elif cwords[0] in subcommands and cwords[0] != 'help':
			subcommand_cls = self.fetch_command(cwords[0])
			parser = subcommand_cls.create_parser('', cwords[0])
			options.extend(
				(min(s_opt.option_strings), s_opt.nargs != 0)
				for s_opt in parser._actions if s_opt.option_strings
			)
			# filter out previously specified options from available options
			prev_opts = {x.split('=')[0] for x in cwords[1:cword - 1]}
			options = (opt for opt in options if opt[0] not in prev_opts)

			# filter options by current input
			options = sorted((k, v) for k, v in options if k.startswith(curr))
			for opt_label, require_arg in options:
				# append '=' to options which require args
				if require_arg:
					opt_label += '='
				print(opt_label)
		# Exit code of the bash completion function is never passed back to
		# the user, so it's safe to always exit with 0.
		# For more details see #25420.
		sys.exit(0)


def execute_from_command_line(argv=None):
	"""Run a ManagementUtility."""
	settings._optional = True
	utility = ManagementUtility(argv)
	utility.execute()



"""
Game Flow Admin methods and functions.
"""
from pyplanet.contrib.command import Command


class FlowAdmin:
	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.admin.app.Admin
		"""
		self.app = app
		self.instance = app.instance

	async def on_start(self):
		await self.instance.permission_manager.register('end_round', 'Force ending a round (warmup, race or custom)', app=self.app, min_level=2)
		await self.instance.permission_manager.register('points_repartition', 'Change the points repartition', app=self.app, min_level=2)

		# Trackmania specific:
		if self.instance.game.game == 'tm' or self.instance.game.game == 'tmnext':
			await self.instance.command_manager.register(
				Command(command='endround', target=self.end_round, perms='admin:end_round', admin=True,
						description='Ends the current round of play.'),
				Command(command='endwuround', target=self.end_wu_round, perms='admin:end_round', admin=True,
						description='Ends the current warm-up round of play.'),
				Command(command='endwu', target=self.end_wu, perms='admin:end_round', admin=True,
						description='Ends the complete warm-up on this map.'),
				Command(command='pointsrepartition', aliases=['pointsrep'], target=self.set_point_repartition,
						perms='admin:points_repartition', admin=True, description='Alters the points repartitioning.')
					.add_param('repartition', nargs='*', type=str, required=True, help='Repartition, comma or space separated.'),
			)

		# Shootmania specific.
		if self.instance.game.game == 'sm':
			pass

	async def end_round(self, player, data, **kwargs):
		await self.instance.gbx.multicall(
			self.instance.gbx('Trackmania.ForceEndRound', encode_json=False, response_id=False),
			self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has forced the current round to end.'.format(player.nickname))
		)

	async def end_wu_round(self, player, data, **kwargs):
		await self.instance.gbx.multicall(
			self.instance.gbx('Trackmania.WarmUp.ForceStopRound', encode_json=False, response_id=False),
			self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has forced the current round to end.'.format(player.nickname))
		)

	async def end_wu(self, player, data, **kwargs):
		await self.instance.gbx.multicall(
			self.instance.gbx('Trackmania.WarmUp.ForceStop', encode_json=False, response_id=False),
			self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has forced the WarmUp to an end.'.format(player.nickname))
		)

	async def set_point_repartition(self, player, data, **kwargs):
		partition = data.repartition

		if len(partition) == 1:
			partition = str(partition[0]).split(',')
		partition = [str(p).strip() for p in partition]

		await self.instance.gbx.multicall(
			self.instance.gbx('Trackmania.SetPointsRepartition', *partition, encode_json=False, response_id=False),
			self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has changed the points repartition to: {}.'.format(
				player.nickname, ','.join(partition))
			)
		)



import asyncio
import async_timeout
import aiohttp
import re

from pyplanet.conf import settings
from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.command import Command
from .view import MusicListView
from .view import PlaylistView


class MusicServer(AppConfig):
	game_dependencies = ['trackmania', 'shootmania']
	app_dependencies = ['core.maniaplanet', 'core.pyplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.lock = asyncio.Lock()
		self.context.signals.listen(mp_signals.map.map_end, self.map_end)
		self.server = None
		self.current_song_index = 0
		self.current_song = None
		self.songs = []
		self.list_view = None
		self.playlist = []
		self.playlist_view = None

	async def on_start(self):
		self.songs = await self.get_songs()
		self.list_view = MusicListView(self)
		self.playlist_view = PlaylistView(self)
		await self.instance.command_manager.register(
			Command(command='play', target=self.play_song, admin=True)
				.add_param(name='songname', type=str, required=True),
			Command(command='song', target=self.get_current_song, admin=False),
			Command(command='songlist', aliases='musiclist', target=self.song_list, admin=False),
			Command(command='playlist', target=self.show_playlist, admin=False),
			Command(command='clearplaylist', target=self.clear_playlist, admin=True),
		)
		self.current_song_index = -1

	async def song_list(self, player, *args, **kwargs):
		self.list_view = MusicListView(self)
		await self.list_view.display(player=player.login)

	async def insert_song(self, player, song):
		self.playlist = self.playlist + [{'player': player, 'song': song}]

	async def show_playlist(self, player, *args, **kwargs):
		self.playlist_view = PlaylistView(self)
		await self.playlist_view.display(player=player.login)

	async def add_to_playlist(self, player, song_index):
		async with self.lock:
			new_song = self.songs[song_index]
			if player.level == 0 and any(item['player'].login == player.login for item in self.playlist):
				message = '$i$f00You already have a song in the playlist! Wait till it\'s been played before adding another.'
				await self.instance.chat(message, player)
				return

			if not any(item['song'] == new_song for item in self.playlist):
				await self.insert_song(player.nickname, new_song)
				message = '$fff{}$z$s$fa0 was added to the playlist by $fff{}$z$s$fa0.'\
					.format(new_song[1]['artist']+" - "+new_song[1]['title'], player.nickname)
				await self.instance.chat(message)
			else:
				message = '$i$f00This song has already been added to the playlist, pick another one.'
				await self.instance.chat(message, player)

	async def drop_from_playlist(self, player, song_info):
		async with self.lock:
			if player.level == 0 and song_info['juke_player'] != player.nickname:
				message = '$i$f00You can only drop your own queued songs!'
				await self.instance.chat(message, player)
			else:
				drop_song = next((item for item in self.playlist if item['song'][1]['title'] == song_info['song_name']), None)
				if drop_song is not None:
					self.playlist.remove(drop_song)
					message = '$fff{}$z$s$fa0 dropped $fff{}$z$s$fa0 from the playlist.'\
						.format(player.nickname, song_info['song_name'])
					await self.instance.chat(message)

	async def map_end(self, *args, **kwargs):
		# Ignore when no songs are added.
		if not self.songs:
			return

		if self.playlist:
			new_song = self.playlist[0]['song']
			self.playlist.pop(0)
		else:
			if self.current_song_index + 2 > len(self.songs):
				new_song = self.songs[0]
				self.current_song_index = 0
			else:
				new_song = self.songs[self.current_song_index + 1]
				self.current_song_index += 1
		try:
			await self.instance.gbx('SetForcedMusic', True, new_song[0])
			self.current_song = new_song
		except Exception as e:
			await self.instance.chat(str(e))

	async def clear_playlist(self, player, data, **kwargs):
		async with self.lock:
			if len(self.playlist) > 0:
				self.playlist.clear()
				message = '$ff0Admin $fff{}$z$s$ff0 has cleared the playlist.'.format(player.nickname)
				await self.instance.chat(message)
			else:
				message = '$i$f00There are currently no songs in the playlist.'
				await self.instance.chat(message, player)

	async def play_song(self, player, data, *args, **kwargs):
		song = str(data.songname)
		try:
			async with aiohttp.ClientSession() as session:
				url, tags = (song, await self.get_tags(session, song))
				self.songs.insert(self.current_song_index + 1, (url, tags))
				message = '$fff{}$z$s$fa0 was added to the songlist by $fff{}$z$s$fa0.'\
					.format(tags['title']+" - "+tags['artist'], player.nickname)
				await self.instance.chat(message)
				await session.close()
		except Exception as e:
			await self.instance.chat(str(e), player)

	async def get_tags(self, session, url):
		with async_timeout.timeout(10):
			async with session.get(url) as response:
				fs = str(await response.content.read(1024))
				tags = {
					'album': 'album',
					'albumartist': 'albumartist',
					'title': 'title',
					'artist': 'artist',
					'date': 'year',
					'tracknumber': 'track',
					'discnumber': 'disc',
					'genre': 'genre'
				}
				for key, value in tags.items():
					if fs.find(key.upper()) > 0:
						end_of_key = fs.find(key.upper()) + len(key) + 1
						end_of_value = fs.find('\\', fs.find(key.upper()))
						value = fs[end_of_key:end_of_value]
						if value.find('(') > 0 or value.find('[') > 0:
							tags[key] = re.sub(r'[^a-zA-Z\d\s\)\]]$', '', value)
						else:
							tags[key] = re.sub(r'[^a-zA-Z\d\s]*$', '', value)
			await response.release()
			return tags

	async def get_songs(self):
		setting = settings.SONGS
		if isinstance(setting, dict) and self.instance.process_name in setting:
			setting = setting[self.instance.process_name]
		if not isinstance(setting, list):
			setting = None

		if not setting:
			message = '$ff0Default song setting not configured in your settings file!'
			await self.instance.chat(message)
			return []

		self.songs.clear()
		songlist = setting
		async with aiohttp.ClientSession() as session:
			tags = [self.get_tags(session, song) for song in songlist]
			tag_list = await asyncio.gather(*tags)
		return [(song.replace("%20", " "), tag_list[i]) for i, song in enumerate(songlist)]

	async def get_current_song(self, player, *args, **kwargs):
		if self.current_song:
			song_url, tags = self.current_song
			message = '$ff0The current song is $fff{}$z$s$ff0 by $fff{}'.format(tags['title'], tags['artist'])
		else:
			message = '$i$f00There is no current song. Skip or restart!'
		await self.instance.chat(message, player)



from peewee import *
from playhouse.migrate import migrate, SchemaMigrator

from pyplanet.apps.core.maniaplanet.models import Map


def upgrade(migrator: SchemaMigrator):
	mx_id = IntegerField(default=None, null=True, index=True)

	migrate(
		migrator.add_column(Map._meta.db_table, 'mx_id', mx_id),
	)


def downgrade(migrator: SchemaMigrator):
	migrate(
		migrator.drop_column(Map._meta.db_table, 'mx_id'),
	)



"""
Maniaplanet Core Models. This models are used in several apps and should be considered as very stable.
"""
from peewee import *
from pyplanet.core.db import TimedModel
from pyplanet.apps.core.maniaplanet.models import Map, Player


class Karma(TimedModel):
	map = ForeignKeyField(Map, index=True)
	"""
	Map on which the player voted.
	"""

	player = ForeignKeyField(Player, index=True)
	"""
	The player who voted.
	"""

	score = IntegerField(
		null=True
	)
	"""
	Karma vote (-1 or 1)
	"""

	expanded_score = FloatField(
		null=True
	)
	"""
	Karma vote (-1, -0.5, 0, 0.5 or 1)
	"""



import logging

from pyplanet.apps.config import AppConfig

from pyplanet.contrib.setting import Setting
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.map.exceptions import ModeIncompatible
from pyplanet.utils.times import format_time
from pyplanet.utils.style import STRIP_ALL, style_strip

logger = logging.getLogger(__name__)


class DynatimeApp(AppConfig):
	game_dependencies = ['trackmania', 'trackmania_next']
	mode_dependencies = ['TimeAttack']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.setting_dynatime_active = Setting(
			'dynatime_active', 'Dynatime Active', Setting.CAT_BEHAVIOUR, type=bool,
			description='Activate Dynamic Round timer based on map bronze time',
			default=True
		)

		self.setting_dynatime_announce = Setting(
			'dynatime_accounce', 'Dynatime Accounce', Setting.CAT_BEHAVIOUR, type=bool,
			description='Announce the current timelimit at the start of each round',
			default=True
		)

		self.setting_dynatime_multiplier = Setting(
			'dynatime_multiplier', 'Dynatime Multiplier', Setting.CAT_BEHAVIOUR, type=int,
			description='Multiply map bronze time by this amount to get round timer',
			default=5
		)

	async def on_init(self):
		await super().on_init()

	async def on_start(self):
		await super().on_start()

		await self.context.setting.register(
			self.setting_dynatime_active,
			self.setting_dynatime_announce,
			self.setting_dynatime_multiplier,
		)

		self.context.signals.listen(mp_signals.map.map_begin, self.on_map_begin)


	async def on_stop(self):
		await super().on_stop()

	async def on_destroy(self):
		await super().on_destroy()

	async def on_map_begin(self, map, **kwargs):
		is_active = await self.setting_dynatime_active.get_value()
		if not is_active:
			return

		multiplier = await self.setting_dynatime_multiplier.get_value(refresh=True)
		mode_settings = await self.instance.mode_manager.get_settings()

		if 'S_TimeLimit' not in mode_settings:
			raise ModeIncompatible('Current mode doesn\'t support Dynatime. Not TimeAttack?')

		timelimit = int(multiplier * map.time_bronze / 1000)
		timelimit_ms = timelimit * 1000

		mode_settings['S_TimeLimit'] = timelimit

		bm_time = format_time(time=map.time_bronze, hide_milliseconds=True)
		new_time = format_time(time=timelimit_ms, hide_milliseconds=True)
		mname = style_strip(map.name, STRIP_ALL)
		chat_message = "$FF0$oDynatime$z$s set new timelimit for map $fff%s$z$ff0 to $fff⏳ %s$ff0, based on the bronze-medal of $fff🏆 %s$z" % ( mname, new_time, bm_time )

		await self.instance.mode_manager.update_settings(mode_settings)

		announce = await self.setting_dynatime_announce.get_value()
		if announce:
			await self.instance.chat(chat_message)
